# Generated by Django 4.2.6 on 2024-01-15 13:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0013_stagedtransactionmodel_bundle_split'),
    ]

    operations = [
        migrations.AddField(
            model_name='ledgermodel',
            name='ledger_xid',
            field=models.SlugField(allow_unicode=True, blank=True, help_text='User Defined Ledger ID', max_length=150, null=True, verbose_name='Ledger Slug'),
        ),
        migrations.AlterUniqueTogether(
            name='ledgermodel',
            unique_together={('entity', 'ledger_xid')},
        ),
    ]



from django.forms import ModelForm, TextInput, Select
from django.utils.translation import gettext_lazy as _

from django_ledger.io.roles import GROUP_INCOME, ASSET_CA_INVENTORY, GROUP_EXPENSES, GROUP_COGS
from django_ledger.models import AccountModel, ItemModel, UnitOfMeasureModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


# #### UNIT OF MEASURES #######
class UnitOfMeasureModelCreateForm(ModelForm):

    def __init__(self, entity_slug: str, user_model, *args, **kwargs):
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        super().__init__(*args, **kwargs)

    class Meta:
        model = UnitOfMeasureModel
        fields = [
            'name',
            'unit_abbr',
            'is_active'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'unit_abbr': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            })
        }


class UnitOfMeasureModelUpdateForm(UnitOfMeasureModelCreateForm):
    pass


# #### PRODUCT ITEMS #######

class ProductCreateForm(ModelForm):
    PRODUCT_OR_SERVICE_ROLES = GROUP_INCOME + GROUP_COGS
    PRODUCT_OR_SERVICE_ROLES.append(ASSET_CA_INVENTORY)

    def __init__(self, entity_slug: str, user_model, *args, **kwargs):
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        super().__init__(*args, **kwargs)

        accounts_qs = AccountModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG,
        ).with_roles(
            roles=self.PRODUCT_OR_SERVICE_ROLES
        ).active()

        # caches the QS for filtering...
        len(accounts_qs)

        self.fields['earnings_account'].queryset = accounts_qs.filter(role__in=GROUP_INCOME)
        self.fields['cogs_account'].queryset = accounts_qs.filter(role__in=GROUP_COGS)
        self.fields['inventory_account'].queryset = accounts_qs.filter(role__in=[ASSET_CA_INVENTORY])

        if 'uom' in self.fields:
            uom_qs = UnitOfMeasureModel.objects.for_entity_active(
                entity_model=self.ENTITY_SLUG
            )
            self.fields['uom'].queryset = uom_qs

        if 'item_type' in self.fields:
            self.fields['item_type'].choices = ItemModel.ITEM_TYPE_CHOICES_PRODUCT

    class Meta:
        model = ItemModel
        fields = [
            'name',
            'sku',
            'upc',
            'item_id',
            'uom',
            'default_amount',
            'earnings_account',
            'cogs_account',
            'inventory_account',
            'sold_as_unit'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'uom': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'earnings_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'cogs_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'inventory_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'sku': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'upc': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_id': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'default_amount': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }
        labels = {
            'name': _('Product Name'),
            'item_type': _('Product Type')
        }

    def clean(self):
        item_model: ItemModel = self.instance
        item_model.item_role = ItemModel.ITEM_ROLE_PRODUCT
        return super().clean()


class ProductUpdateForm(ProductCreateForm):
    pass


# #### SERVICE ITEMS #######

class ServiceCreateForm(ModelForm):
    SERVICE_ROLES = GROUP_INCOME + GROUP_COGS

    def __init__(self, entity_slug: str, user_model, *args, **kwargs):
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        super().__init__(*args, **kwargs)

        accounts_qs = AccountModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG
        ).with_roles(
            roles=self.SERVICE_ROLES,
        ).active()

        # caches the QS for filtering...
        len(accounts_qs)

        self.fields['earnings_account'].queryset = accounts_qs.filter(role__in=GROUP_INCOME)
        self.fields['cogs_account'].queryset = accounts_qs.filter(role__in=GROUP_COGS)

        if 'uom' in self.fields:
            uom_qs = UnitOfMeasureModel.objects.for_entity_active(
                entity_model=self.ENTITY_SLUG,
            )
            self.fields['uom'].queryset = uom_qs

    class Meta:
        model = ItemModel
        fields = [
            'name',
            'sku',
            'upc',
            'item_id',
            'uom',
            'default_amount',
            'earnings_account',
            'cogs_account',
            'sold_as_unit'
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'uom': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'earnings_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'cogs_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'inventory_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'sku': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'upc': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_id': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'default_amount': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }
        labels = {
            'name': _('Service Name'),
            'sold_as_unit': _('Sold as Unit?')
        }

    def clean(self):
        item_model: ItemModel = self.instance
        item_model.item_role = ItemModel.ITEM_ROLE_SERVICE
        return super().clean()


class ServiceUpdateForm(ServiceCreateForm):
    pass


# #### EXPENSE ITEMS #######
class ExpenseItemCreateForm(ModelForm):
    def __init__(self, entity_slug: str, user_model, *args, **kwargs):
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        super().__init__(*args, **kwargs)

        accounts_qs = AccountModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG
        ).with_roles(
            roles=GROUP_EXPENSES
        ).active()

        self.fields['expense_account'].queryset = accounts_qs.filter(role__in=GROUP_EXPENSES)

        if 'uom' in self.fields:
            uom_qs = UnitOfMeasureModel.objects.for_entity(
                entity_model=self.ENTITY_SLUG
            ).for_user(user_model=self.USER_MODEL)
            self.fields['uom'].queryset = uom_qs

    class Meta:
        model = ItemModel
        fields = [
            'name',
            'upc',
            'sku',
            'uom',
            'item_type',
            'default_amount',
            'expense_account',
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('The item name...')
            }),
            'uom': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'expense_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'sku': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'upc': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                'placeholder': _('The UPC code of the item, if any...')
            }),
            'item_id': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'default_amount': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }
        labels = {
            'name': _('Expense Name')
        }

    def clean(self):
        item_model: ItemModel = self.instance
        item_model.item_role = ItemModel.ITEM_ROLE_EXPENSE
        return super().clean()


class ExpenseItemUpdateForm(ExpenseItemCreateForm):
    class Meta(ExpenseItemCreateForm.Meta):
        fields = [
            'name',
            'upc',
            'sku',
            'default_amount',
            'expense_account',
            'is_active'
        ]


# #### INVENTORY ITEMS #######
class InventoryItemCreateForm(ModelForm):

    def __init__(self, entity_slug: str, user_model, *args, **kwargs):
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        super().__init__(*args, **kwargs)

        inventory_account_qs = AccountModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG,
        ).with_roles(
            roles=[ASSET_CA_INVENTORY]
        ).active()
        self.fields['inventory_account'].queryset = inventory_account_qs

        if 'uom' in self.fields:
            uom_qs = UnitOfMeasureModel.objects.for_entity_active(
                entity_model=self.ENTITY_SLUG,
            )
            self.fields['uom'].queryset = uom_qs

    class Meta:
        model = ItemModel
        fields = [
            'name',
            'uom',
            'upc',
            'sku',
            'item_id',
            'item_type',
            'default_amount',
            'inventory_account',
        ]
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'uom': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'inventory_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'earnings_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'cogs_account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'upc': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'sku': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'item_id': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'default_amount': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }
        labels = {
            'name': _('Inventory Name')
        }

    def clean(self):
        item_model: ItemModel = self.instance
        item_model.item_role = ItemModel.ITEM_ROLE_INVENTORY
        return super().clean()


class InventoryItemUpdateForm(InventoryItemCreateForm):
    class Meta(InventoryItemCreateForm.Meta):
        fields = [
            'name',
            'upc',
            'sku',
            'item_id',
            'default_amount',
            'inventory_account',
        ]



# Generated by Django 4.2 on 2023-04-20 20:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0006_alter_journalentrymodel_je_number_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='customermodel',
            name='django_ledg_custome_16f95a_idx',
        ),
        migrations.AlterUniqueTogether(
            name='accountmodel',
            unique_together={('coa_model', 'code')},
        ),
        migrations.AddField(
            model_name='accountmodel',
            name='role_default',
            field=models.BooleanField(blank=True, null=True, verbose_name='Coa Role Default Account'),
        ),
        migrations.AlterField(
            model_name='accountmodel',
            name='role',
            field=models.CharField(choices=[('Assets', (('asset_ca_cash', 'Current Asset'), ('asset_ca_mkt_sec', 'Marketable Securities'), ('asset_ca_recv', 'Receivables'), ('asset_ca_inv', 'Inventory'), ('asset_ca_uncoll', 'Uncollectibles'), ('asset_ca_prepaid', 'Prepaid'), ('asset_ca_other', 'Other Liquid Assets'), ('asset_lti_notes', 'Notes Receivable'), ('asset_lti_land', 'Land'), ('asset_lti_sec', 'Securities'), ('asset_ppe_build', 'Buildings'), ('asset_ppe_build_accum_depr', 'Buildings - Accum. Depreciation'), ('asset_ppe_plant', 'Plant'), ('asset_ppe_plant_depr', 'Plant - Accum. Depreciation'), ('asset_ppe_equip', 'Equipment'), ('asset_ppe_equip_accum_depr', 'Equipment - Accum. Depreciation'), ('asset_ia', 'Intangible Assets'), ('asset_ia_accum_amort', 'Intangible Assets - Accum. Amortization'), ('asset_adjustment', 'Other Assets'))), ('Liabilities', (('lia_cl_acc_payable', 'Accounts Payable'), ('lia_cl_wages_payable', 'Wages Payable'), ('lia_cl_int_payable', 'Interest Payable'), ('lia_cl_taxes_payable', 'Taxes Payable'), ('lia_cl_st_notes_payable', 'Short Term Notes Payable'), ('lia_cl_ltd_mat', 'Current Maturities of Long Tern Debt'), ('lia_cl_def_rev', 'Deferred Revenue'), ('lia_cl_other', 'Other Liabilities'), ('lia_ltl_notes', 'Long Term Notes Payable'), ('lia_ltl_bonds', 'Bonds Payable'), ('lia_ltl_mortgage', 'Mortgage Payable'))), ('Equity', (('eq_capital', 'Capital'), ('eq_stock_common', 'Common Stock'), ('eq_stock_preferred', 'Preferred Stock'), ('eq_adjustment', 'Other Equity Adjustments'), ('eq_dividends', 'Dividends & Distributions to Shareholders'), ('in_operational', 'Operational Income'), ('in_passive', 'Investing/Passive Income'), ('in_interest', 'Interest Income'), ('in_gain_loss', 'Capital Gain/Loss Income'), ('in_other', 'Other Income'), ('cogs_regular', 'Cost of Goods Sold'), ('ex_regular', 'Regular Expense'), ('ex_interest', 'Interest Expense'), ('ex_taxes', 'Tax Expense'), ('ex_capital', 'Capital Expense'), ('ex_depreciation', 'Depreciation Expense'), ('ex_amortization', 'Amortization Expense'), ('ex_other', 'Other Expense'))), ('Root', (('root_coa', 'CoA Root Account'), ('root_assets', 'Assets Root Account'), ('root_liabilities', 'Liabilities Root Account'), ('root_capital', 'Capital Root Account'), ('root_income', 'Income Root Account'), ('root_cogs', 'COGS Root Account'), ('root_expenses', 'Expenses Root Account')))], max_length=30, verbose_name='Account Role'),
        ),
        migrations.AlterUniqueTogether(
            name='accountmodel',
            unique_together={('coa_model', 'role', 'role_default'), ('coa_model', 'code')},
        ),
        migrations.AddIndex(
            model_name='customermodel',
            index=models.Index(fields=['entity_model', 'customer_number'], name='django_ledg_entity__7b3ad8_idx'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['entity_model', 'vendor_number'], name='django_ledg_entity__266aa0_idx'),
        ),
        migrations.AddIndex(
            model_name='vendormodel',
            index=models.Index(fields=['vendor_number'], name='django_ledg_vendor__5f9f46_idx'),
        ),
    ]



from datetime import date, timedelta, datetime, time
from decimal import Decimal
from itertools import cycle
from logging import getLogger, DEBUG
from random import randint, choice
from typing import Optional, Literal
from zoneinfo import ZoneInfo

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.test import TestCase
from django.test.client import Client
from django.utils.timezone import get_default_timezone

from django_ledger.io.io_generator import EntityDataGenerator
from django_ledger.models import JournalEntryModel, LedgerModel, TransactionModel, AccountModel, AccountModelQuerySet
from django_ledger.models.entity import EntityModel, EntityModelQuerySet, UserModel

UserModel = get_user_model()


class DjangoLedgerBaseTest(TestCase):
    FY_STARTS = None
    CAPITAL_CONTRIBUTION = None
    START_DATE = None
    DAYS_FORWARD = 9 * 30
    TX_QUANTITY = 50
    user_model = None
    TEST_DATA = list()
    CLIENT = None
    TZ = None
    N = 1
    USER_EMAIL = None
    PASSWORD = None
    USERNAME = None
    logger = None
    accrual_cycle = cycle([True, False])

    @classmethod
    def setUpTestData(cls):

        cls.logger = getLogger(__name__)
        cls.logger.setLevel(level=DEBUG)

        cls.USERNAME: str = 'testuser'
        cls.PASSWORD: str = 'NeverUseThisPassword12345'
        cls.USER_EMAIL: str = 'testuser@djangoledger.com'

        cls.DAYS_FWD: int = randint(180, 180 * 3)
        cls.TZ = get_default_timezone()
        cls.START_DATE = cls.get_random_date(as_datetime=True)

        cls.CLIENT = Client(enforce_csrf_checks=False)

        try:
            cls.user_model = UserModel.objects.get(username=cls.USERNAME)
        except ObjectDoesNotExist:
            cls.user_model = UserModel.objects.create_user(
                username=cls.USERNAME,
                password=cls.PASSWORD,
                email=cls.USER_EMAIL,
            )

        cls.FY_STARTS = list(str(i) for i in range(1, 13))
        cls.TEST_DATA = list()
        cls.CAPITAL_CONTRIBUTION = Decimal('50000.00')
        cls.ENTITY_MODEL_QUERYSET: Optional[EntityModelQuerySet] = None

        cls.create_entity_models(n=cls.N)
        cls.populate_entity_models()

    @classmethod
    def get_random_date(cls, as_datetime: bool = False) -> date:
        dt = date(
            year=choice(range(1990, 2020)),
            month=choice(range(1, 13)),
            day=choice(range(1, 28))
        )
        if as_datetime:
            if not settings.USE_TZ:
                return datetime.combine(
                    dt,
                    time(
                        hour=randint(1, 23),
                        minute=randint(1, 59)
                    ),
                )
            return datetime.combine(
                dt,
                time(
                    hour=randint(1, 23),
                    minute=randint(1, 59)
                ),
                tzinfo=ZoneInfo(settings.TIME_ZONE)
            )
        return dt

    @classmethod
    def login_client(cls):
        # cls.logger.info('Logging in client...')
        cls.CLIENT.login(
            username=cls.USERNAME,
            password=cls.PASSWORD
        )

    @classmethod
    def logout_client(cls):
        # cls.logger.info('Logging out client...')
        cls.CLIENT.logout()

    @classmethod
    def refresh_test_data(cls, n: int = None):
        N = n if n else cls.N
        cls.TEST_DATA = [cls.get_random_entity_data() for _ in range(N)]

    @classmethod
    def get_random_entity_data(cls) -> dict:
        return {
            'slug': f'a-cool-slug-{randint(10000, 99999)}',
            'name': f'Testing Inc-{randint(100000, 999999)}',
            'address_1': f'{randint(100000, 999999)} Main St',
            'address_2': f'Suite {randint(1000, 9999)}',
            'city': 'Charlotte',
            'state': 'NC',
            'zip_code': '28202',
            'country': 'US',
            'email': 'mytest@testinginc.com',
            'website': 'http://www.mytestingco.com',
            'fy_start_month': choice(cls.FY_STARTS),
            'admin': cls.user_model,
            'accrual_method': next(cls.accrual_cycle)
        }

    def get_random_entity_model(self) -> EntityModel:
        if self.ENTITY_MODEL_QUERYSET:
            return choice(self.ENTITY_MODEL_QUERYSET)
        raise ValueError('EntityModels have not been populated.')

    def create_entity_model(self, use_accrual_method: bool = False, fy_start_month: int = 1) -> EntityModel:
        """
        Creates a new blank EntityModel for testing purposes.

        Parameters
        ----------
        use_accrual_method: bool
            Whether to use the accrual method. Defaults to False.
        fy_start_month:
            The month to start the financial year. Defaults to 1 (January).

        Returns
        -------
        EntityModel
        """
        return EntityModel.create_entity(
            name='Testing Inc-{randint(100000, 999999)',
            use_accrual_method=use_accrual_method,
            fy_start_month=fy_start_month,
            admin=self.user_model
        )

    @classmethod
    def create_entity_models(cls, save=True, n: int = 5):
        cls.refresh_test_data(n)
        for ent_data in cls.TEST_DATA:
            entity_model = EntityModel.add_root(**ent_data)
            entity_model.admin = cls.user_model
            entity_model.clean()
            if save:
                entity_model.save()

    @classmethod
    def populate_entity_models(cls):
        entities_qs = EntityModel.objects.all()
        for entity_model in entities_qs:
            data_generator = EntityDataGenerator(
                user_model=cls.user_model,
                entity_model=entity_model,
                start_dttm=cls.START_DATE,
                capital_contribution=cls.CAPITAL_CONTRIBUTION,
                days_forward=cls.DAYS_FORWARD,
                tx_quantity=cls.TX_QUANTITY
            )
            cls.logger.info(f'Populating Entity {entity_model.name}...')
            data_generator.populate_entity()
        cls.ENTITY_MODEL_QUERYSET = entities_qs

    def get_random_draft_date(self):
        return self.START_DATE + timedelta(days=randint(0, 365))

    def get_random_account(self,
                           entity_model: EntityModel,
                           balance_type: Literal['credit', 'debit', None] = None,
                           active: bool = True,
                           locked: bool = False) -> AccountModel:
        """
        Returns 1 random AccountModel with the specified balance_type.
        """
        account_qs: AccountModelQuerySet = entity_model.get_coa_accounts(active=active, locked=locked)
        account_qs = account_qs.filter(balance_type=balance_type) if balance_type else account_qs
        return choice(account_qs)

    def get_random_ledger(self,
                          entity_model: EntityModel,
                          qs_limit: int = 100,
                          posted: bool = True) -> LedgerModel:
        """
        Returns 1 random LedgerModel object.
        """
        ledger_model_qs = entity_model.get_ledgers(
            posted=posted).filter(
            journal_entries__count__gt=0
        )

        # no need to check because data generator will always populate an entity with sample data.
        # if not ledger_model.exists():
        #     for i in range(3):
        #         LedgerModel.objects.create(
        #             name=f"{i}Example Ledger {randint(10000, 99999)}",
        #             ledger_xid=f"{i}example-ledger-xid-{randint(10000, 99999)}",
        #             entity=entity_model,
        #         )

        # limits the queryset in case of large querysets...
        return choice(ledger_model_qs[:qs_limit])

    def get_random_je(self,
                      entity_model: EntityModel,
                      ledger_model: Optional[LedgerModel] = None,
                      posted: bool = True,
                      qs_limit: int = 100
                      ) -> JournalEntryModel:
        """.
        Returns 1 random JournalEntryModel object.
        """
        if not ledger_model:
            ledger_model: LedgerModel = self.get_random_ledger(
                entity_model=entity_model,
                qs_limit=qs_limit,
            )
        else:
            entity_model.validate_ledger_model_for_entity(ledger_model)
        journal_entry_qs = ledger_model.journal_entries.all()

        # no need to check because data generator will always populate an entity with sample data.
        # if not je_model.exists():
        #     for i in range(3):
        #         random_je_activity = choice([category[0] for category in JournalEntryModel.ACTIVITIES])
        #         JournalEntryModel.objects.create(
        #             je_number=f"{i}example-je-num-{randint(10000, 99999)}",
        #             description=f"{i}Random Journal Entry Desc {randint(10000, 99999)}",
        #             is_closing_entry=False,
        #             activity=random_je_activity,
        #             posted=False,
        #             locked=False,
        #             ledger=ledger_model,
        #         )

        if posted:
            journal_entry_qs = journal_entry_qs.posted()

        journal_entry_qs = journal_entry_qs.select_related('ledger', 'ledger__entity')
        # limits the queryset in case of large querysets...
        return choice(journal_entry_qs[:qs_limit])

    def get_random_transaction(self,
                               entity_model: EntityModel,
                               je_model: Optional[JournalEntryModel] = None,
                               posted: bool = True,
                               qs_limit: int = 100) -> TransactionModel:
        """
        Returns all TransactionModel related to a random or specified JournalEntryModel.
        """
        if not je_model:
            je_model = self.get_random_je(entity_model=entity_model, posted=posted)
        else:
            ledger_model = je_model.ledger
            entity_model.validate_ledger_model_for_entity(ledger_model)
        txs_model_qs = je_model.transactionmodel_set.all()
        return choice(txs_model_qs[:qs_limit])

    def resolve_url_patterns(self, url_patterns):
        self.URL_PATTERNS = {
            p.name: set(p.pattern.converters.keys()) for p in url_patterns
        }

    def resolve_url_kwars(self):
        url_patterns = getattr(self, 'URL_PATTERNS', None)
        if not url_patterns:
            raise ValidationError(
                message='Must call resolve_url_patterns before calling resolve_url_kwars.'
            )

        return set.union(*[
            set(v) for v in self.URL_PATTERNS.values()
        ])



# Generated by Django 4.1.5 on 2023-02-04 16:10

from decimal import Decimal
import django.core.validators
from django.db import migrations, models
import django_ledger.models.accounts


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0002_alter_journalentrymodel_managers_and_more'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='accountmodel',
            name='django_ledg_role_1bff96_idx',
        ),
        migrations.AlterField(
            model_name='accountmodel',
            name='code',
            field=models.CharField(max_length=10, validators=[django_ledger.models.accounts.account_code_validator], verbose_name='Account Code'),
        ),
        migrations.AlterField(
            model_name='accountmodel',
            name='role',
            field=models.CharField(choices=[('Assets', (('asset_ca_cash', 'Current Asset'), ('asset_ca_mkt_sec', 'Marketable Securities'), ('asset_ca_recv', 'Receivables'), ('asset_ca_inv', 'Inventory'), ('asset_ca_uncoll', 'Uncollectibles'), ('asset_ca_prepaid', 'Prepaid'), ('asset_ca_other', 'Other Liquid Assets'), ('asset_lti_notes', 'Notes Receivable'), ('asset_lti_land', 'Land'), ('asset_lti_sec', 'Securities'), ('asset_ppe_build', 'Buildings'), ('asset_ppe_build_accum_depr', 'Buildings - Accum. Depreciation'), ('asset_ppe_plant', 'Plant'), ('asset_ppe_plant_depr', 'Plant - Accum. Depreciation'), ('asset_ppe_equip', 'Equipment'), ('asset_ppe_equip_accum_depr', 'Equipment - Accum. Depreciation'), ('asset_ia', 'Intangible Assets'), ('asset_ia_accum_amort', 'Intangible Assets - Accum. Amortization'), ('asset_adjustment', 'Other Assets'))), ('Liabilities', (('lia_cl_acc_payable', 'Accounts Payable'), ('lia_cl_wages_payable', 'Wages Payable'), ('lia_cl_int_payable', 'Interest Payable'), ('lia_cl_taxes_payable', 'Taxes Payable'), ('lia_cl_st_notes_payable', 'Notes Payable'), ('lia_cl_ltd_mat', 'Current Maturities of Long Tern Debt'), ('lia_cl_def_rev', 'Deferred Revenue'), ('lia_cl_other', 'Other Liabilities'), ('lia_ltl_notes', 'Notes Payable'), ('lia_ltl_bonds', 'Bonds Payable'), ('lia_ltl_mortgage', 'Mortgage Payable'))), ('Equity', (('eq_capital', 'Capital'), ('eq_stock_common', 'Common Stock'), ('eq_stock_preferred', 'Preferred Stock'), ('eq_adjustment', 'Other Equity Adjustments'), ('eq_dividends', 'Dividends & Distributions to Shareholders'), ('in_operational', 'Operational Income'), ('in_passive', 'Investing/Passive Income'), ('in_interest', 'Interest Income'), ('in_gain_loss', 'Capital Gain/Loss Income'), ('in_other', 'Other Income'), ('cogs_regular', 'Cost of Goods Sold'), ('ex_regular', 'Regular Expense'), ('ex_interest', 'Interest Expense'), ('ex_taxes', 'Tax Expense'), ('ex_capital', 'Capital Expense'), ('ex_depreciation', 'Depreciation Expense'), ('ex_amortization', 'Amortization Expense'), ('ex_other', 'Other Expense'))), ('Root', (('root_coa', 'CoA Root Account'), ('root_assets', 'Assets Root Account'), ('root_liabilities', 'Liabilities Root Account'), ('root_capital', 'Capital Root Account'), ('root_income', 'Income Root Account'), ('root_cogs', 'COGS Root Account'), ('root_expenses', 'Expenses Root Account')))], max_length=30, verbose_name='Account Role'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='equipment_estimate',
            field=models.DecimalField(decimal_places=2, default=Decimal('0.00'), help_text='Estimated equipment cost to complete the quoted work.', max_digits=20, validators=[django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Equipment Cost Estimate'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='labor_estimate',
            field=models.DecimalField(decimal_places=2, default=Decimal('0.00'), help_text='Estimated labor cost to complete the quoted work.', max_digits=20, validators=[django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Labor Cost of labor estimate'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='material_estimate',
            field=models.DecimalField(decimal_places=2, default=0.0, help_text='Estimated material cost to complete the quoted work.', max_digits=20, validators=[django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Material Cost Estimate'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='other_estimate',
            field=models.DecimalField(decimal_places=2, default=Decimal('0.00'), help_text='Estimated equipment cost to complete the quoted work.', max_digits=20, validators=[django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Other Cost Estimate'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='revenue_estimate',
            field=models.DecimalField(decimal_places=2, default=Decimal('0.00'), help_text='Estimated cost to complete the quoted work.', max_digits=20, validators=[django.core.validators.MinValueValidator(limit_value=0)], verbose_name='Total revenue estimate'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='status',
            field=models.CharField(choices=[('draft', 'Draft'), ('in_review', 'In Review'), ('approved', 'Approved'), ('completed', 'Completed'), ('void', 'Void'), ('canceled', 'Canceled')], default='draft', max_length=10, verbose_name='Estimate Model Status'),
        ),
        migrations.AlterField(
            model_name='estimatemodel',
            name='title',
            field=models.CharField(max_length=250, validators=[django.core.validators.MinLengthValidator(limit_value=5, message='EstimateModel Title length must be greater than 5')], verbose_name='Customer Estimate Title'),
        ),
        migrations.AddIndex(
            model_name='accountmodel',
            index=models.Index(fields=['locked'], name='django_ledg_locked_c02b15_idx'),
        ),
    ]



# Generated by Django 4.1.5 on 2023-02-10 12:56

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('django_ledger', '0003_remove_accountmodel_django_ledg_role_1bff96_idx_and_more'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='itemmodel',
            name='depth',
        ),
        migrations.RemoveField(
            model_name='itemmodel',
            name='numchild',
        ),
        migrations.RemoveField(
            model_name='itemmodel',
            name='path',
        ),
    ]



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""
from typing import Optional

from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django.views.generic import ListView, CreateView, UpdateView, DeleteView

from django_ledger.forms.customer import CustomerModelForm
from django_ledger.models.customer import CustomerModel, CustomerModelQueryset
from django_ledger.models.entity import EntityModel
from django_ledger.views.mixins import DjangoLedgerSecurityMixIn


class CustomerModelModelViewQuerySetMixIn(DjangoLedgerSecurityMixIn):
    queryset: Optional[CustomerModelQueryset] = None

    def get_queryset(self):
        if self.queryset is None:
            self.queryset = CustomerModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug'],
            ).order_by('-updated')
        return self.queryset


class CustomerModelListView(CustomerModelModelViewQuerySetMixIn, ListView):
    template_name = 'django_ledger/customer/customer_list.html'
    PAGE_TITLE = _('Customer List')
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'dashicons:businesswoman'
    }
    context_object_name = 'customers'


class CustomerModelCreateView(CustomerModelModelViewQuerySetMixIn, CreateView):
    template_name = 'django_ledger/customer/customer_create.html'
    PAGE_TITLE = _('Create New Customer')
    form_class = CustomerModelForm
    context_object_name = 'customer'
    extra_context = {
        'page_title': PAGE_TITLE,
        'header_title': PAGE_TITLE,
        'header_subtitle_icon': 'dashicons:businesswoman'
    }

    def get_success_url(self):
        return reverse('django_ledger:customer-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })

    def form_valid(self, form):
        customer_model: CustomerModel = form.save(commit=False)
        entity_model = EntityModel.objects.for_user(
            user_model=self.request.user
        ).get(slug__exact=self.kwargs['entity_slug'])
        customer_model.entity_model = entity_model
        customer_model.save()
        return super().form_valid(form)


class CustomerModelUpdateView(CustomerModelModelViewQuerySetMixIn, UpdateView):
    template_name = 'django_ledger/customer/customer_update.html'
    PAGE_TITLE = _('Customer Update')
    form_class = CustomerModelForm
    context_object_name = 'customer'
    slug_url_kwarg = 'customer_pk'
    slug_field = 'uuid'

    def get_context_data(self, **kwargs):
        context = super(CustomerModelUpdateView, self).get_context_data(**kwargs)
        customer_model: CustomerModel = self.object
        context['page_title'] = self.PAGE_TITLE
        context['header_title'] = self.PAGE_TITLE
        context['header_subtitle'] = customer_model.customer_number
        context['header_subtitle_icon'] = 'dashicons:businesswoman'
        return context

    def get_success_url(self):
        return reverse('django_ledger:customer-list',
                       kwargs={
                           'entity_slug': self.kwargs['entity_slug']
                       })

    def form_valid(self, form):
        form.save()
        return super().form_valid(form)


class CustomerModelDeleteView(CustomerModelModelViewQuerySetMixIn, DeleteView):
    pass



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    - Miguel Sanda <msanda@arrobalytics.com>
    - Michael Noel <noel.michael87@gmail.com>
"""

from django.forms import ModelForm, modelformset_factory, BaseModelFormSet, TextInput, Select, ValidationError
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import check_tx_balance
from django_ledger.models import EntityModel
from django_ledger.models.journal_entry import JournalEntryModel
from django_ledger.models.transactions import TransactionModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class TransactionModelForm(ModelForm):
    class Meta:
        model = TransactionModel
        fields = [
            'account',
            'tx_type',
            'amount',
            'description'
        ]
        widgets = {
            'account': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'tx_type': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'amount': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'description': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
        }


class TransactionModelFormSet(BaseModelFormSet):

    def __init__(self, *args, entity_model: EntityModel, je_model: JournalEntryModel, **kwargs):
        super().__init__(*args, **kwargs)
        je_model.validate_for_entity(entity_model)
        self.JE_MODEL: JournalEntryModel = je_model
        self.ENTITY_MODEL = entity_model

        account_qs = self.ENTITY_MODEL.get_coa_accounts().active().order_by('code')

        for form in self.forms:
            form.fields['account'].queryset = account_qs
            if self.JE_MODEL.is_locked():
                form.fields['account'].disabled = True
                form.fields['tx_type'].disabled = True
                form.fields['amount'].disabled = True

    def get_queryset(self):
        return self.JE_MODEL.transactionmodel_set.all()

    def clean(self):
        if any(self.errors):
            return
        for form in self.forms:
            if self.can_delete and self._should_delete_form(form):
                continue
        txs_balances = [{
            'tx_type': tx.cleaned_data.get('tx_type'),
            'amount': tx.cleaned_data.get('amount')
        } for tx in self.forms if not self._should_delete_form(tx)]
        balance_ok = check_tx_balance(txs_balances, perform_correction=False)
        if not balance_ok:
            raise ValidationError(message=_('Credits and Debits do not balance.'))


def get_transactionmodel_formset_class(journal_entry_model: JournalEntryModel):
    can_delete = not journal_entry_model.is_locked()
    return modelformset_factory(
        model=TransactionModel,
        form=TransactionModelForm,
        formset=TransactionModelFormSet,
        can_delete=can_delete,
        extra=6 if can_delete else 0)



from random import randint

from django.forms import ModelForm, TextInput, Textarea, HiddenInput
from django.utils.translation import gettext_lazy as _

from django_ledger.models.chart_of_accounts import ChartOfAccountModel
from django_ledger.models.entity import EntityModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class ChartOfAccountsModelCreateForm(ModelForm):
    FORM_ID_SEP = '___'

    def __init__(self, entity_model: EntityModel, *args, **kwargs):
        self.ENTITY_MODEL = entity_model
        super().__init__(*args, **kwargs)
        self.fields['entity'].disabled = True
        self.fields['entity'].required = True
        self.form_id: str = self.get_form_id()

    def clean_entity(self):
        return self.ENTITY_MODEL

    def get_form_id(self) -> str:
        return f'coa-model-create-form-{self.ENTITY_MODEL.slug}{self.FORM_ID_SEP}{randint(100000, 999999)}'

    class Meta:
        model = ChartOfAccountModel
        fields = [
            'entity',
            'name',
            'description'
        ]
        labels = {
            'name': _('Name'),
            'description': _('Description'),
        }
        widgets = {
            'entity': HiddenInput(),
            'name': TextInput(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
                }),
            'description': Textarea(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
                }
            ),
        }


class ChartOfAccountsModelUpdateForm(ModelForm):

    FORM_ID_SEP = '___'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.form_id: str = self.get_form_id()


    def get_form_id(self) -> str:
        instance: ChartOfAccountModel = self.instance
        return f'coa-model-update-form-{instance.slug}{self.FORM_ID_SEP}{randint(100000, 999999)}'

    class Meta:
        model = ChartOfAccountModel
        fields = [
            'name',
            'active'
        ]
        labels = {
            'name': _('Name'),
            'description': _('Description'),
        }
        widgets = {
            'name': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'description': Textarea(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
        }



from django import forms
from django.forms import (
    ModelForm, BaseModelFormSet, modelformset_factory,
    Select, NumberInput, HiddenInput,
    TextInput, ValidationError
)
from django.utils.translation import gettext_lazy as _

from django_ledger.models import (
    StagedTransactionModel,
    ImportJobModel,
    EntityModel
)
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class ImportJobModelCreateForm(ModelForm):

    def __init__(self, entity_model: EntityModel, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_MODEL: EntityModel = entity_model
        self.fields['bank_account_model'].queryset = self.ENTITY_MODEL.bankaccountmodel_set.all().active()

    ofx_file = forms.FileField(
        label='Select File...',
        required=True,
        widget=forms.FileInput(
            attrs={
                'class': 'file-input',
                'accept': '.ofx,.qfx'
            })
    )

    class Meta:
        model = ImportJobModel
        fields = [
            'bank_account_model',
            'description',
        ]
        widgets = {
            'description': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                'placeholder': _('Name this import...')
            }),
            'bank_account_model': Select(
                attrs={
                    'class': DJANGO_LEDGER_FORM_INPUT_CLASSES,
                }),
        }
        help_texts = {
            'bank_account_model': _('Select the bank account to import transactions from.'),
        }


class ImportJobModelUpdateForm(ModelForm):
    class Meta:
        model = ImportJobModel
        fields = [
            'description'
        ]
        widgets = {
            'description': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            })
        }


class StagedTransactionModelForm(ModelForm):
    tx_import = forms.BooleanField(initial=False, required=False)
    tx_split = forms.BooleanField(initial=False, required=False)

    def __init__(self, base_formset_instance, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.BASE_FORMSET_INSTANCE = base_formset_instance

        # for IDE typing hints...
        instance: StagedTransactionModel = getattr(self, 'instance', None)

        # avoids multiple DB queries rendering the formset...
        self.fields['unit_model'].choices = self.BASE_FORMSET_INSTANCE.UNIT_MODEL_CHOICES

        # avoids multiple DB queries rendering the formset...
        self.fields['account_model'].choices = self.BASE_FORMSET_INSTANCE.ACCOUNT_MODEL_CHOICES

        if instance:
            if not instance.is_children():
                self.fields['amount_split'].widget = HiddenInput()
                self.fields['amount_split'].disabled = True
            else:
                self.fields['bundle_split'].widget = HiddenInput()
                self.fields['bundle_split'].disabled = True

            if not instance.can_have_account():
                self.fields['account_model'].widget = HiddenInput()
                self.fields['account_model'].disabled = True

            if not instance.can_have_unit():
                self.fields['unit_model'].widget = HiddenInput()
                self.fields['unit_model'].disabled = True

            if not instance.can_import():
                self.fields['tx_import'].widget = HiddenInput()
                self.fields['tx_import'].disabled = True
            if not instance.can_split():
                self.fields['tx_split'].widget = HiddenInput()
                self.fields['tx_split'].disabled = True

    def clean_account_model(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if staged_txs_model.has_children():
            return None
        return self.cleaned_data['account_model']

    def clean_unit_model(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if not staged_txs_model.can_have_unit():
            if staged_txs_model.parent_id:
                return staged_txs_model.parent.unit_model
        return self.cleaned_data['unit_model']

    def clean_tx_import(self):
        staged_txs_model: StagedTransactionModel = self.instance
        if staged_txs_model.is_children():
            return False
        return self.cleaned_data['tx_import']

    def clean(self):
        # cannot import and split at the same time
        if self.cleaned_data['tx_import'] and self.cleaned_data['tx_split']:
            raise ValidationError(message=_('Cannot import and split at the same time'))

    class Meta:
        model = StagedTransactionModel
        fields = [
            'tx_import',
            'bundle_split',
            'amount_split',
            'account_model',
            'unit_model'
        ]
        widgets = {
            'account_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'unit_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'amount_split': NumberInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small'
            })
        }


class BaseStagedTransactionModelFormSet(BaseModelFormSet):

    def __init__(self, *args,
                 entity_model: EntityModel,
                 import_job_model: ImportJobModel,
                 **kwargs):
        super().__init__(*args, **kwargs)

        # validates that the job import model belongs to the entity model...
        if import_job_model.entity_uuid != entity_model.uuid:
            raise ValidationError(message=_('Import job does not belong to this entity'))

        self.ENTITY_MODEL = entity_model
        self.IMPORT_JOB_MODEL: ImportJobModel = import_job_model
        self.MAPPED_ACCOUNT_MODEL = self.IMPORT_JOB_MODEL.bank_account_model.account_model

        self.account_model_qs = entity_model.get_coa_accounts().available().exclude(
            uuid__exact=self.MAPPED_ACCOUNT_MODEL.uuid
        )
        self.ACCOUNT_MODEL_CHOICES = [
            (a.uuid, a) if i > 0 else (None, '----') for i, a in
            enumerate(self.account_model_qs)
        ]

        self.unit_model_qs = entity_model.entityunitmodel_set.all()
        self.UNIT_MODEL_CHOICES = [
            (u.uuid, u) if i > 0 else (None, '----') for i, u in
            enumerate(self.unit_model_qs)
        ]
        self.queryset = self.IMPORT_JOB_MODEL.stagedtransactionmodel_set.all().is_pending()

    def get_form_kwargs(self, index):
        return {
            'base_formset_instance': self,
        }


StagedTransactionModelFormSet = modelformset_factory(
    model=StagedTransactionModel,
    form=StagedTransactionModelForm,
    formset=BaseStagedTransactionModelFormSet,
    can_delete=True,
    can_order=False,
    extra=0
)



from decimal import Decimal
from random import choice, randint

from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist, ValidationError
from django.db.models import Count
from django.db.utils import IntegrityError

from django_ledger.forms.transactions import (
    get_transactionmodel_formset_class,
    TransactionModelForm,
    TransactionModelFormSet
)
from django_ledger.io.io_core import get_localdate
from django_ledger.models import (
    TransactionModel, EntityModel, AccountModel, LedgerModel, JournalEntryModel,
    TransactionModelValidationError, JournalEntryValidationError
)
from django_ledger.tests.base import DjangoLedgerBaseTest

UserModel = get_user_model()


class TransactionModelTest(DjangoLedgerBaseTest):

    def test_invalid_balance(self):
        entity_model = self.get_random_entity_model()
        txs_model = self.get_random_transaction(entity_model=entity_model)

        # transaction model does not allow for negative balances
        txs_model.amount = Decimal('-100.00')

        with self.assertRaises(ValidationError):
            txs_model.full_clean()


class TransactionModelFormTest(DjangoLedgerBaseTest):

    def test_valid_data(self):
        entity_model: EntityModel = self.get_random_entity_model()

        account_model = str(self.get_random_account(entity_model=entity_model, balance_type='credit').uuid),
        random_tx_type = choice([tx_type[0] for tx_type in TransactionModel.TX_TYPE])

        form_data = {
            'account': account_model[0],
            'tx_type': random_tx_type,
            'amount': Decimal(randint(10000, 99999)),
            'description': 'Bought Something ...'
        }

        form = TransactionModelForm(form_data)

        self.assertTrue(form.is_valid(), msg=f'Form is invalid with error: {form.errors}')
        with self.assertRaises(IntegrityError):
            form.save()

    def test_invalid_tx_type(self):
        account_model = choice(AccountModel.objects.filter(balance_type='credit'))
        form = TransactionModelForm({
            'account': account_model,
            'tx_type': 'crebit patty',
        })
        self.assertFalse(form.is_valid(), msg='tx_type other than credit / debit shouldn\'t be valid')

    def test_blank_data(self):
        form = TransactionModelForm()
        self.assertFalse(form.is_valid(), msg='Form without data is supposed to be invalid')

    def test_invalid_account(self):
        with self.assertRaises(ObjectDoesNotExist):
            form = TransactionModelForm({
                'account': 'Asset',
            })
            form.is_valid()


class TransactionModelFormSetTest(DjangoLedgerBaseTest):

    def get_random_txs_formsets(self,
                                entity_model: EntityModel,
                                ledger_model: LedgerModel = None,
                                je_model: JournalEntryModel = None) -> TransactionModelFormSet:
        """
        Returns a TransactionModelFormSet with prefilled form data.
        """

        if ledger_model:
            # if ledger model provided, get a je_model from provided ledger model...
            je_model: JournalEntryModel = self.get_random_je(
                entity_model=entity_model,
                ledger_model=ledger_model
            ) if not je_model else je_model

        else:

            # get a journal entry that has transactions...
            je_model = JournalEntryModel.objects.for_entity(
                entity_model=entity_model
            ).annotate(
                txs_count=Count('transactionmodel')).filter(
                txs_count__gt=0).order_by('-timestamp').first()

        TransactionModelFormSet = get_transactionmodel_formset_class(journal_entry_model=je_model)

        txs_formset = TransactionModelFormSet(
            entity_slug=entity_model.slug,
            user_model=self.user_model,
            je_model=je_model,
            ledger_pk=je_model.ledger_id,
        )
        return txs_formset

    def test_valid_formset(self):
        """
        Saved Transaction instances should have identical detail with initial formset.
        """
        entity_model: EntityModel = self.get_random_entity_model()
        ledger_model: LedgerModel = self.get_random_ledger(entity_model=entity_model)
        je_model: JournalEntryModel = self.get_random_je(entity_model=entity_model, ledger_model=ledger_model)
        credit_account: AccountModel = self.get_random_account(entity_model=entity_model, balance_type='credit')
        debit_account: AccountModel = self.get_random_account(entity_model=entity_model, balance_type='debit')
        transaction_amount = Decimal.from_float(randint(10000, 99999))

        txs_formset = self.get_random_txs_formsets(
            entity_model=entity_model,
            je_model=je_model,
            ledger_model=ledger_model
        )

        self.assertTrue(
            txs_formset.is_valid(),
            msg=f"Formset is not valid, error: {txs_formset.errors}")

        txs_instances = txs_formset.save(commit=False)
        for txs in txs_instances:
            if not txs.journal_entry_id:
                txs.journal_entry_id = je_model.uuid

        txs_instances = txs_formset.save()
        for txs in txs_instances:
            if txs.tx_type == 'credit':
                self.assertEqual(
                    txs.account, credit_account,
                    msg=f'Saved Transaction record has mismatched Credit Account from the submitted formset. Saved:{txs.account} | form:{credit_account}')

            elif txs.tx_type == 'debit':
                self.assertEqual(
                    txs.account, debit_account,
                    msg=f'Saved Transaction record has mismatched Debit Account from the submitted formset. Saved:{txs.account} | form:{debit_account}')

            self.assertEqual(
                txs.amount, Decimal(transaction_amount),
                msg=f'Saved Transaction record has mismatched total amount from the submitted formset. Saved:{txs.amount} | form:{transaction_amount}')

    def test_imbalance_transactions(self):
        """
        Imbalanced Transactions should be invalid.
        """
        entity_model: EntityModel = self.get_random_entity_model()

        txs_formset = self.get_random_txs_formsets(entity_model=entity_model)

        self.assertFalse(
            txs_formset.is_valid(),
            msg=f"Formset is supposed to be invalid because of imbalance transaction"
        )

    def test_je_locked(self):
        """
        Transaction on locked a locked Journal Entry should fail.
        """
        entity_model: EntityModel = self.get_random_entity_model()
        ledger_model: LedgerModel = self.get_random_ledger(
            entity_model=entity_model
        )

        je_model: JournalEntryModel = self.get_random_je(
            entity_model=entity_model,
            ledger_model=ledger_model
        )
        je_model.mark_as_locked(commit=True, raise_exception=False)

        self.assertTrue(je_model.is_locked())
        txs_model = je_model.transactionmodel_set.all().first()
        txs_model.amount += Decimal.from_float(1.00)

        with self.assertRaises(TransactionModelValidationError):
            txs_model.save()

        with self.assertRaises(
                TransactionModelValidationError,
                msg=f'Cannot create transaction on locked Journal Entry'
        ):
            je_model.transactionmodel_set.create(
                amount=Decimal.from_float(100.00),
                account=self.get_random_account(entity_model=entity_model, balance_type='debit')
            )

    def test_ledger_lock(self):
        """
        Transaction on locked a locked Ledger should fail.
        """
        entity_model: EntityModel = self.get_random_entity_model()
        ledger_model = self.get_random_ledger(entity_model=entity_model)
        ledger_model.post(commit=True, raise_exception=False)
        self.assertTrue(ledger_model.is_posted())
        ledger_model.lock(commit=True, raise_exception=False)
        self.assertTrue(ledger_model.is_locked())

        with self.assertRaises(
                JournalEntryValidationError,
                msg='Cannot create Journal Entries on locked ledgers.'
        ):
            ledger_model.journal_entries.create(
                timestamp=get_localdate(),
                description='Test Journal Entry'
            )

        je_model = ledger_model.journal_entries.first()

        with self.assertRaises(
                JournalEntryValidationError,
                msg='Cannot unpost journal entry on locked ledgers'
        ):
            je_model.mark_as_unposted(commit=True, raise_exception=True)


class GetTransactionModelFormSetClassTest(DjangoLedgerBaseTest):

    def test_unlocked_journal_entry_formset(self):
        """
        The Formset will contain 6 extra forms & delete fields if Journal Entry is unlocked.
        """
        entity_model: EntityModel = self.get_random_entity_model()
        ledger_model: LedgerModel = self.get_random_ledger(entity_model=entity_model)
        je_model: JournalEntryModel = self.get_random_je(entity_model=entity_model, ledger_model=ledger_model)
        je_model.mark_as_unlocked(commit=True)

        transaction_model_form_set = get_transactionmodel_formset_class(journal_entry_model=je_model)
        txs_formset = transaction_model_form_set(
            user_model=self.user_model,
            je_model=je_model,
            ledger_pk=ledger_model,
            entity_slug=entity_model.slug,
        )

        self.assertTrue(not je_model.is_locked(),
                        msg="At this point in this test case, Journal Entry should be unlocked.")

        delete_field = '<input type="checkbox" name="form-0-DELETE" id="id_form-0-DELETE">'
        self.assertInHTML(
            delete_field,
            txs_formset.as_table(),
            msg_prefix='Transactions Formset with unlocked Journal Entry should have `can_delete` enabled'
        )

        self.assertEqual(len(txs_formset), 6,
                         msg='Transactions Formset with unlocked Journal Entry should have 6 extras')

    def test_locked_journal_entry_formset(self):
        """
        The Formset will contain no extra forms & only forms with Transaction if Journal Entry is locked.
        """
        entity_model: EntityModel = self.get_random_entity_model()
        ledger_model: LedgerModel = self.get_random_ledger(entity_model=entity_model)
        je_model: JournalEntryModel = self.get_random_je(entity_model=entity_model, ledger_model=ledger_model)
        # transaction_pairs = randint(1, 12)
        # self.get_random_transactions(entity_model=entity_model, je_model=je_model,
        #                              pairs=transaction_pairs)  # Fill Journal Entry with Transactions

        je_model.mark_as_locked(commit=True)
        self.assertTrue(
            je_model.is_locked(),
            msg="Journal Entry should be locked in this test case")

        transaction_model_form_set = get_transactionmodel_formset_class(journal_entry_model=je_model)

        txs_formset = transaction_model_form_set(
            user_model=self.user_model,
            je_model=je_model,
            ledger_pk=ledger_model,
            entity_slug=entity_model.slug,
            queryset=je_model.transactionmodel_set.all().order_by('account__code')
        )

        self.assertEqual(
            len(txs_formset), (je_model.transactionmodel_set.count()),  # Convert pairs to total count
            msg="Transactions Formset with unlocked Journal Entry did not match the expected count")



from datetime import timedelta
from uuid import uuid4

from django.contrib.admin import TabularInline, ModelAdmin
from django.db.models import Count
from django.forms import BaseInlineFormSet
from django.urls import reverse
from django.utils.html import format_html

from django_ledger.admin.chart_of_accounts import ChartOfAccountsInLine
from django_ledger.io.io_core import get_localtime
from django_ledger.models import EntityUnitModel
from django_ledger.models.entity import EntityModel, EntityManagementModel


class EntityManagementInLine(TabularInline):
    model = EntityManagementModel
    extra = 1
    fields = [
        'user'
    ]


class EntityUnitModelInLineFormSet(BaseInlineFormSet):

    def save_new(self, form, commit=True):
        setattr(form.instance, self.fk.name, self.instance)
        if commit:
            unit_model = EntityUnitModel.add_root(
                instance=super().save_new(form, commit=False)
            )
            return unit_model
        return super().save_new(form, commit=False)


class EntityUnitModelInLine(TabularInline):
    model = EntityUnitModel
    formset = EntityUnitModelInLineFormSet
    extra = 0
    readonly_fields = [
        'slug'
    ]
    fields = [
        'slug',
        'name',
        'document_prefix',
        'active',
        'hidden'
    ]


class EntityModelAdmin(ModelAdmin):
    list_display = [
        'slug',
        'name',
        'admin',
        'accrual_method',
        'last_closing_date',
        'hidden',
        'get_coa_count',
        'add_ledger_link',
        'dashboard_link',
        'balance_sheet_link',
        'income_statement_link',
        'cash_flow_statement_link'
    ]
    readonly_fields = [
        'depth',
        'path',
        'numchild',
        'last_closing_date',
        'default_coa'
    ]
    fieldsets = [
        (
            'Entity Information', {
                'fields': [
                    'name',
                    'admin',
                    'fy_start_month',
                    'accrual_method',
                    'hidden',
                    'picture'
                ]
            }
        ),
        (
            'Contact Information', {
                'fields': [
                    'address_1',
                    'address_2',
                    'city',
                    'state',
                    'zip_code',
                    'email',
                    'website',
                    'phone'
                ]
            }
        ),
        (
            'Chart of Accounts', {
                'fields': [
                    'default_coa'
                ]
            }
        )
    ]
    inlines = [
        ChartOfAccountsInLine,
        EntityUnitModelInLine,
        EntityManagementInLine
    ]
    actions = [
        'add_code_of_accounts',
        'populate_random_data'
    ]

    class Meta:
        model = EntityModel

    def get_queryset(self, request):
        qs = super().get_queryset(request=request)
        qs = qs.annotate(Count('chartofaccountmodel'))
        if request.user.is_superuser:
            return qs
        return qs.for_user(user_model=request.user)

    def add_ledger_link(self, obj):
        add_ledger_url = reverse('admin:django_ledger_ledgermodel_add')
        return format_html('<a class="addlink" href="{url}?entity_slug={slug}">Add Ledger</a>',
                           url=add_ledger_url,
                           slug=obj.slug)

    def balance_sheet_link(self, obj: EntityModel):
        add_ledger_url = reverse(
            viewname='django_ledger:entity-bs',
            kwargs={
                'entity_slug': obj.slug
            })
        return format_html('<a class="viewlink" href="{url}">View</a>',
                           url=add_ledger_url,
                           slug=obj.slug)

    balance_sheet_link.short_description = 'Balance Sheet'

    def income_statement_link(self, obj: EntityModel):
        add_ledger_url = reverse(
            viewname='django_ledger:entity-ic',
            kwargs={
                'entity_slug': obj.slug
            })
        return format_html('<a class="viewlink" href="{url}">View</a>',
                           url=add_ledger_url,
                           slug=obj.slug)

    income_statement_link.short_description = 'P&L'

    def cash_flow_statement_link(self, obj: EntityModel):
        add_ledger_url = reverse(
            viewname='django_ledger:entity-cf',
            kwargs={
                'entity_slug': obj.slug
            })
        return format_html('<a class="viewlink" href="{url}">View</a>',
                           url=add_ledger_url,
                           slug=obj.slug)

    cash_flow_statement_link.short_description = 'Cash Flow'

    def dashboard_link(self, obj: EntityModel):
        add_ledger_url = reverse(
            viewname='django_ledger:entity-dashboard',
            kwargs={
                'entity_slug': obj.slug
            })
        return format_html('<a class="viewlink" href="{url}">View</a>',
                           url=add_ledger_url,
                           slug=obj.slug)

    dashboard_link.short_description = 'Dashboard'

    def add_code_of_accounts(self, request, queryset):
        lt = get_localtime().isoformat()
        for entity_model in queryset:
            entity_model.create_chart_of_accounts(
                coa_name=f'{entity_model.name} CoA {lt}',
                commit=True,
                assign_as_default=False
            )

    def populate_random_data(self, request, queryset):
        for entity_model in queryset:
            start_date = get_localtime() - timedelta(days=180)
            entity_model.populate_random_data(
                start_date=start_date,
                days_forward=180,
                tx_quantity=25)

    def get_coa_count(self, obj):
        return obj.chartofaccountmodel__count

    get_coa_count.short_description = 'CoA Count'

    def save_model(self, request, obj, form, change):
        if not change:
            if obj.uuid is None:
                obj.uuid = uuid4()
            EntityModel.add_root(instance=obj)
            return
        super().save_model(request, obj, form, change)

    def has_delete_permission(self, request, obj=None):
        return False



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
    * Miguel Sanda <msanda@arrobalytics.com>
"""

from calendar import monthrange
from datetime import timedelta, date
from typing import Tuple, Optional

from django.contrib.auth.mixins import PermissionRequiredMixin, LoginRequiredMixin
from django.core.exceptions import ValidationError, ObjectDoesNotExist, ImproperlyConfigured
from django.db.models import Q
from django.http import Http404, HttpResponse, HttpResponseNotFound
from django.urls import reverse
from django.utils.dateparse import parse_date
from django.utils.translation import gettext_lazy as _
from django.views.generic.dates import YearMixin, MonthMixin, DayMixin

from django_ledger.models import EntityModel, InvoiceModel, BillModel, LedgerModel
from django_ledger.models.entity import EntityModelFiscalPeriodMixIn
from django_ledger.settings import DJANGO_LEDGER_AUTHORIZED_SUPERUSER


class ContextFromToDateMixin:
    FROM_DATE_CONTEXT_NAME = 'from_date'
    TO_DATE_CONTEXT_NAME = 'to_date'

    def get_from_date_context_name(self) -> str:
        return self.FROM_DATE_CONTEXT_NAME

    def get_to_date_context_name(self) -> str:
        return self.TO_DATE_CONTEXT_NAME


class YearlyReportMixIn(YearMixin, ContextFromToDateMixin, EntityModelFiscalPeriodMixIn):

    def get_from_date(self, year: int = None, fy_start: int = None, **kwargs) -> date:
        return self.get_year_start_date(year, fy_start)

    def get_to_date(self, year: int = None, fy_start: int = None, **kwargs) -> date:
        return self.get_year_end_date(year, fy_start)

    def get_from_to_dates(self, year: int = None, fy_start: int = None, **kwargs) -> Tuple[date, date]:
        from_date = self.get_from_date(year, fy_start, **kwargs)
        to_date = self.get_to_date(year, fy_start, **kwargs)
        return from_date, to_date

    def get_year_start_date(self, year: int = None, fy_start: int = None) -> date:
        if not year:
            year = self.get_year()
        return self.get_fy_start(year, fy_start)

    def get_year_end_date(self, year: int = None, fy_start: int = None) -> date:
        if not year:
            year = self.get_year()
        return self.get_fy_end(year, fy_start)

    def get_context_data(self, **kwargs):
        context = super(YearlyReportMixIn, self).get_context_data(**kwargs)
        year = self.get_year()
        context['year'] = year
        context['next_year'] = year + 1
        context['previous_year'] = year - 1
        year_start = self.get_year_start_date(year)
        year_end = self.get_year_end_date(year)
        context['year_start'] = year_start
        context['year_end'] = year_end

        if self.get_from_date_context_name() not in context:
            context[self.get_from_date_context_name()] = year_start
        if self.get_to_date_context_name() not in context:
            context[self.get_to_date_context_name()] = year_end

        context['has_year'] = True
        return context


class QuarterlyReportMixIn(YearMixin, ContextFromToDateMixin, EntityModelFiscalPeriodMixIn):
    quarter = None
    quarter_url_kwarg = 'quarter'

    def parse_quarter(self, quarter) -> int:
        try:
            if not isinstance(quarter, int):
                quarter = int(quarter)
            try:
                self.validate_quarter(quarter)
            except ValidationError:
                raise Http404(_("Invalid quarter number"))
        except ValueError:
            raise Http404(_(f"Invalid quarter format. Cannot parse {quarter} into integer."))
        return quarter

    def get_quarter(self) -> int:
        quarter = self.quarter
        if quarter is None:
            try:
                quarter = self.kwargs[self.quarter_url_kwarg]
            except KeyError:
                try:
                    quarter = self.request.GET[self.quarter_url_kwarg]
                except KeyError:
                    raise Http404(_("No quarter specified"))
        quarter = self.parse_quarter(quarter)
        return quarter

    def get_from_date(self, quarter: int = None, year: int = None, fy_start: int = None, **kwargs) -> date:
        return self.get_quarter_start_date(quarter, year, fy_start)

    def get_to_date(self, quarter: int = None, year: int = None, fy_start: int = None, **kwargs) -> date:
        return self.get_quarter_end_date(quarter, year, fy_start)

    def get_from_to_dates(self,
                          quarter: int = None,
                          year: int = None,
                          fy_start: int = None,
                          **kwargs) -> Tuple[date, date]:
        from_date = self.get_from_date(quarter=quarter, year=year, fy_start=fy_start, **kwargs)
        to_date = self.get_to_date(quarter=quarter, year=year, fy_start=fy_start, **kwargs)
        return from_date, to_date

    def get_quarter_start_date(self, quarter: int = None, year: int = None, fy_start: int = None) -> date:
        if not year:
            year = self.get_year()
        if not quarter:
            quarter = self.get_quarter()
        return self.get_quarter_start(year, quarter, fy_start)

    def get_quarter_end_date(self, quarter: int = None, year: int = None, fy_start: int = None) -> date:
        if not year:
            year = self.get_year()
        if not quarter:
            quarter = self.get_quarter()
        return self.get_quarter_end(year, quarter, fy_start)

    def get_context_data(self, **kwargs) -> dict:
        context = super(QuarterlyReportMixIn, self).get_context_data(**kwargs)
        quarter = self.get_quarter()
        year = self.get_year()
        context['quarter'] = quarter
        context['next_quarter'] = self.get_next_quarter(quarter)
        context['previous_quarter'] = self.get_previous_quarter(quarter)
        quarter_start = self.get_quarter_start_date(year=year, quarter=quarter)
        quarter_end = self.get_quarter_end_date(year=year, quarter=quarter)
        context['quarter_start'] = quarter_start
        context['quarter_end'] = quarter_end
        context[self.get_from_date_context_name()] = quarter_start
        context[self.get_to_date_context_name()] = quarter_end
        context['has_quarter'] = True
        return context

    def get_next_quarter(self, quarter) -> int:
        if quarter != 4:
            return quarter + 1

    def get_previous_quarter(self, quarter) -> int:
        if quarter != 1:
            return quarter - 1


class MonthlyReportMixIn(YearlyReportMixIn, ContextFromToDateMixin, MonthMixin):

    def get_from_date(self, month: int = None, year: int = None, **kwargs) -> date:
        return self.get_month_start_date(month=month, year=year)

    def get_to_date(self, month: int = None, year: int = None, **kwargs) -> date:
        return self.get_month_end_date(month=month, year=year)

    def get_from_to_dates(self,
                          month: int = None,
                          year: int = None,
                          **kwargs) -> Tuple[date, date]:
        from_date = self.get_from_date(month=month, year=year, **kwargs)
        to_date = self.get_to_date(month=month, year=year, **kwargs)
        return from_date, to_date

    def get_month_start_date(self, month: int = None, year: int = None) -> date:
        if not month:
            month = int(self.get_month())
        if not year:
            year = self.get_year()
        return date(year=year, month=month, day=1)

    def get_month_end_date(self, month: int = None, year: int = None) -> date:
        if not month:
            month = int(self.get_month())
        if not year:
            year = self.get_year()
        last_day = monthrange(year, month)[1]
        return date(year=year, month=month, day=last_day)

    def get_next_month(self, month) -> int:
        if month != 12:
            return month + 1
        return 1

    def get_previous_month(self, month) -> int:
        if month != 1:
            return month - 1
        return 12

    def get_context_data(self, **kwargs):
        context = super(MonthlyReportMixIn, self).get_context_data(**kwargs)
        month = int(self.get_month())
        year = int(self.get_year())
        context['month'] = month
        context['next_month'] = self.get_next_month(month)
        context['previous_month'] = self.get_previous_month(month)
        month_start = self.get_month_start_date(year=year, month=month)
        month_end = self.get_month_end_date(year=year, month=month)
        context['month_start'] = month_start
        context['month_end'] = month_end
        context[self.get_from_date_context_name()] = month_start
        context[self.get_to_date_context_name()] = month_end
        context['has_month'] = True
        return context


class DateReportMixIn(MonthlyReportMixIn, ContextFromToDateMixin, DayMixin):

    def get_context_data(self, **kwargs):
        context = super(MonthlyReportMixIn, self).get_context_data(**kwargs)
        view_date = self.get_date()
        context['has_date'] = True
        context['next_day'] = view_date + timedelta(days=1)
        context['previous_day'] = view_date - timedelta(days=1)
        context['view_date'] = view_date
        context[self.get_from_date_context_name()] = view_date
        context[self.get_to_date_context_name()] = view_date
        return context

    def get_date(self) -> date:
        return date(
            year=self.get_year(),
            month=self.get_month(),
            day=self.get_day()
        )

    def get_from_date(self, month: int = None, year: int = None, **kwargs) -> date:
        return self.get_date()

    def get_to_date(self, month: int = None, year: int = None, **kwargs) -> date:
        return self.get_date()

    def get_from_to_dates(self, month: int = None, year: int = None, **kwargs) -> Tuple[date, date]:
        dt = self.get_from_date(month=month, year=year, **kwargs)
        return dt, dt


# todo: need to incorporate in base view...
class FromToDatesParseMixIn:
    DJL_FROM_DATE_PARAM: str = 'from_date'
    DJL_TO_DATE_PARAM: str = 'to_date'
    DJL_NO_FROM_DATE_RAISE_404: bool = True
    DJL_NO_TO_DATE_RAISE_404: bool = True

    def get_from_date(self, query_param: str = None) -> date:
        if not query_param:
            query_param = self.DJL_FROM_DATE_PARAM
        parsed_date = self.parse_date_from_query_param(query_param)
        if not parsed_date and self.DJL_NO_FROM_DATE_RAISE_404:
            raise Http404(_(f'Must provide {query_param} date parameter.'))
        return parsed_date

    def get_to_date(self, query_param: str = None) -> date:
        if not query_param:
            query_param = self.DJL_TO_DATE_PARAM
        parsed_date = self.parse_date_from_query_param(query_param)
        if not parsed_date and self.DJL_NO_TO_DATE_RAISE_404:
            raise Http404(_(f'Must provide {query_param} date parameter.'))
        return parsed_date

    def get_from_to_dates(self, query_param: str = None) -> Tuple[date, date]:
        from_date = self.get_from_date(query_param)
        to_date = self.get_to_date(query_param)
        return from_date, to_date

    def parse_date_from_query_param(self, query_param: str):
        param_date = self.request.GET.get(query_param)
        if param_date:
            parsed_date = parse_date(param_date)
            if not parsed_date:
                raise Http404(_(f'Invalid {query_param} {param_date} provided'))
            param_date = parsed_date
        return param_date


class SuccessUrlNextMixIn:

    def has_next_url(self):
        return self.request.GET.get('next') is not None

    def get_success_url(self):
        next = self.request.GET.get('next')
        if next:
            return next
        return reverse('django_ledger:home')


class DjangoLedgerSecurityMixIn(LoginRequiredMixin, PermissionRequiredMixin):
    ENTITY_SLUG_URL_KWARG = 'entity_slug'
    ENTITY_MODEL_CONTEXT_NAME = 'entity_model'
    AUTHORIZE_SUPERUSER: bool = DJANGO_LEDGER_AUTHORIZED_SUPERUSER
    permission_required = []

    def __init__(self, *args, **kwargs):
        self.AUTHORIZED_ENTITY_MODEL: Optional[EntityModel] = None
        super().__init__(*args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context[self.ENTITY_MODEL_CONTEXT_NAME] = self.get_authorized_entity_instance(raise_exception=False)
        return context

    def get_login_url(self):
        return reverse('django_ledger:login')

    def get_entity_slug(self):
        return self.kwargs[self.ENTITY_SLUG_URL_KWARG]

    def get_entity_slug_kwarg(self):
        if self.ENTITY_SLUG_URL_KWARG is None:
            raise ImproperlyConfigured(
                _('ENTITY_SLUG_URL_KWARG must be provided.')
            )
        return self.ENTITY_SLUG_URL_KWARG

    def get_superuser_authorization(self):
        return self.AUTHORIZE_SUPERUSER

    def has_permission(self):
        has_perm = super().has_permission()
        if not has_perm:
            return False

        entity_slug_kwarg = self.get_entity_slug_kwarg()
        entity_model_qs = self.get_authorized_entity_queryset()

        if self.request.user.is_authenticated:
            if entity_slug_kwarg in self.kwargs:
                try:
                    self.AUTHORIZED_ENTITY_MODEL = entity_model_qs.get(slug__exact=self.kwargs[entity_slug_kwarg])
                except ObjectDoesNotExist:
                    return False
            return True
        return False

    def get_authorized_entity_queryset(self):
        return EntityModel.objects.for_user(
            user_model=self.request.user,
            authorized_superuser=self.get_superuser_authorization(),
        )

    def get_authorized_entity_instance(self, raise_exception: bool = True) -> Optional[EntityModel]:
        if self.AUTHORIZED_ENTITY_MODEL is None:
            if raise_exception:
                raise Http404()
            return
        return self.AUTHORIZED_ENTITY_MODEL

    def get_authorized_entity_instance_name(self) -> Optional[str]:
        entity_model: EntityModel = self.get_authorized_entity_instance()
        if not entity_model:
            return None
        return entity_model.name


class EntityUnitMixIn:
    UNIT_SLUG_KWARG = 'unit_slug'
    UNIT_SLUG_QUERY_PARAM = 'unit'

    def get_unit_slug(self):
        unit_slug = self.kwargs.get(self.UNIT_SLUG_KWARG)
        if not unit_slug:
            unit_slug = self.request.GET.get(self.UNIT_SLUG_QUERY_PARAM)
        return unit_slug

    def get_context_data(self, **kwargs):
        context = super(EntityUnitMixIn, self).get_context_data(**kwargs)
        unit_slug = self.get_unit_slug()
        context['unit_slug'] = unit_slug

        by_unit = any([
            True if unit_slug else False,
            self.request.GET.get('by_unit') is not None
        ])

        context['by_unit'] = by_unit
        return context


class DigestContextMixIn:
    IO_DIGEST_UNBOUNDED = False
    IO_DIGEST_BOUNDED = False

    IO_DIGEST_UNBOUNDED_CONTEXT_NAME = 'tx_digest'
    IO_MANAGER_UNBOUNDED_CONTEXT_NAME = 'tx_digest_context'

    IO_DIGEST_BOUNDED_CONTEXT_NAME = 'equity_digest'
    IO_MANAGER_BOUNDED_CONTEXT_NAME = 'equity_digest_context'

    def get_io_digest_unbounded_context_name(self):
        return self.IO_DIGEST_UNBOUNDED_CONTEXT_NAME

    def get_io_manager_unbounded_context_name(self):
        return self.IO_MANAGER_UNBOUNDED_CONTEXT_NAME

    def get_io_digest_bounded_context_name(self):
        return self.IO_DIGEST_BOUNDED_CONTEXT_NAME

    def get_io_manager_bounded_context_name(self):
        return self.IO_MANAGER_BOUNDED_CONTEXT_NAME

    def get_context_data(self, **kwargs):
        context = super(DigestContextMixIn, self).get_context_data(**kwargs)
        return self.get_io_digest(context=context, **kwargs)

    def get_io_digest(self,
                      context,
                      from_date=None,
                      to_date=None,
                      **kwargs):

        if any([
            self.IO_DIGEST_UNBOUNDED,
            self.IO_DIGEST_BOUNDED
        ]):

            by_period = self.request.GET.get('by_period')
            io_model: EntityModel | LedgerModel = self.object
            if not to_date:
                to_date = context['to_date']
            if not from_date:
                from_date = context['from_date']

            # gets the unit if view inherits EntityUnitMixIn...
            if hasattr(self, 'get_unit_slug'):
                unit_slug = getattr(self, 'get_unit_slug')()
            else:
                unit_slug = None

            if self.IO_DIGEST_UNBOUNDED:
                io_digest = io_model.digest(
                    user_model=self.request.user,
                    entity_slug=io_model.entity_slug if isinstance(io_model, LedgerModel) else None,
                    to_date=to_date,
                    unit_slug=unit_slug,
                    by_period=True if by_period else False,
                    process_ratios=True,
                    process_roles=True,
                    process_groups=True
                )

                context[self.get_io_manager_unbounded_context_name()] = io_digest
                context[self.get_io_digest_unbounded_context_name()] = io_digest.get_io_data()

            if self.IO_DIGEST_BOUNDED:
                io_digest_equity = io_model.digest(
                    user_model=self.request.user,
                    equity_only=True,
                    to_date=to_date,
                    from_date=from_date,
                    unit_slug=unit_slug,
                    by_period=True if by_period else False,
                    process_ratios=True,
                    process_roles=True,
                    process_groups=True
                )

                context[self.get_io_manager_bounded_context_name()] = io_digest_equity
                context[self.get_io_digest_bounded_context_name()] = io_digest_equity.get_io_data()

            # todo: how is this used??....
            context['date_filter'] = to_date
        return context


class UnpaidElementsMixIn:
    FETCH_UNPAID_INVOICES: bool = False
    FETCH_UNPAID_BILLS: bool = False

    def get_context_data(self, **kwargs):
        context = super(UnpaidElementsMixIn, self).get_context_data(**kwargs)
        context['invoices'] = self.get_unpaid_invoices_qs(context)
        context['bills'] = self.get_unpaid_bills_qs(context)
        return context

    def get_unpaid_invoices_qs(self, context, from_date=None, to_date=None):
        if self.FETCH_UNPAID_INVOICES:
            from_date = context['from_date'] if not from_date else from_date
            to_date = context['to_date'] if not to_date else to_date

            qs = InvoiceModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug']
            ).for_user(
                user_model=self.request.user
            ).approved().filter(
                Q(date_approved__gte=from_date) &
                Q(date_approved__lte=to_date)
            ).select_related('customer').order_by('date_due')

            unit_slug = self.get_unit_slug()
            if unit_slug:
                qs = qs.filter(ledger__journal_entries__entity_unit__slug__exact=unit_slug)

            return qs

    def get_unpaid_bills_qs(self, context, from_date=None, to_date=None):
        if self.FETCH_UNPAID_BILLS:
            from_date = context['from_date'] if not from_date else from_date
            to_date = context['to_date'] if not to_date else to_date

            qs = BillModel.objects.for_entity(
                entity_model=self.kwargs['entity_slug']
            ).for_user(
                user_model=self.request.user
            ).unpaid().filter(
                Q(date_approved__gte=from_date) &
                Q(date_approved__lte=to_date)
            ).select_related('vendor').order_by('date_due')

            unit_slug = self.get_unit_slug()

            if unit_slug:
                qs = qs.filter(ledger__journal_entries__entity_unit__slug__exact=unit_slug)

            return qs


class BaseDateNavigationUrlMixIn:
    BASE_DATE_URL_KWARGS = (
        'entity_slug',
        'unit_slug',
        'ledger_pk',
        'account_pk',
        'coa_slug'
    )

    def get_context_data(self, **kwargs):
        context = super(BaseDateNavigationUrlMixIn, self).get_context_data(**kwargs)
        self.get_base_date_nav_url(context)
        return context

    def get_base_date_nav_url(self, context, **kwargs):
        view_name = context['view'].request.resolver_match.url_name
        view_name_base = '-'.join(view_name.split('-')[:2])
        context['date_navigation_url'] = reverse(
            f'django_ledger:{view_name_base}',
            kwargs={
                k: v for k, v in self.kwargs.items() if
                k in self.BASE_DATE_URL_KWARGS
            })


class PDFReportMixIn:
    class PDFReportEnum:
        BS = 'BS'
        IS = 'IS'
        CFS = 'CFS'

    pdf_report_enum = PDFReportEnum
    pdf_report_type: Optional[PDFReportEnum] = None
    pdf_format_query_param = 'format'
    pdf_format_query_param_value = 'pdf'
    pdf_subtitle_query_param = 'report_subtitle'
    pdf_io_mixin_function_map = {
        PDFReportEnum.BS: 'get_balance_sheet_statement',
        PDFReportEnum.IS: 'get_income_statement',
        PDFReportEnum.CFS: 'get_cash_flow_statement',
    }

    def get_pdf_func_name(self):
        if not self.pdf_report_type:
            raise NotImplementedError(f'Must define pdf_report_type from {self.PDFReportEnum.__name__}')
        return self.pdf_io_mixin_function_map[self.pdf_report_type]

    def get_pdf(self):
        self.object = self.get_object()
        io_model = self.object
        pdf_func_name = self.get_pdf_func_name()
        pdf = getattr(io_model, pdf_func_name)(
            entity_slug=self.kwargs.get('entity_slug'),
            from_date=self.get_pdf_from_date(),
            to_date=self.get_pdf_to_date(),
            user_model=self.request.user,
            subtitle=self.get_pdf_subtitle()
        )
        pdf.create_pdf_report()
        return pdf

    def get_pdf_subtitle(self) -> str:
        return self.request.GET.get(self.pdf_subtitle_query_param)

    def get_pdf_from_date(self) -> Optional[date]:
        ctx = getattr(self, 'get_context_data')()
        return ctx['from_date']

    def get_pdf_to_date(self) -> date:
        ctx = getattr(self, 'get_context_data')()
        return ctx['to_date']

    def get_pdf_response(self) -> HttpResponse:
        pdf = self.get_pdf()
        response = HttpResponse(
            bytes(pdf.output()),
            content_type="application/pdf",
        )
        response.headers['Content-Disposition'] = f'attachment; filename={pdf.get_pdf_filename()}'
        return response

    def get(self, request, **kwargs):
        if request.GET.get(self.pdf_format_query_param) == self.pdf_format_query_param_value:
            return self.get_pdf_response()
        return super().get(request, **kwargs)



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

The EstimateModel provides the means to estimate customer requests, jobs or quotes that may ultimately be considered
contracts, if approved. The EstimateModels will estimate revenues and costs associated with a specific scope of work
which is documented using ItemTransactionModels.

Once approved, the user may initiate purchase orders, bills and invoices that will be associated with the EstimateModel
for tracking purposes. It is however not required to always have an EstimateModel, but recommended in order to be able
to produce more specific financial reports associated with a specific scope of work.
"""
import warnings
from datetime import date
from decimal import Decimal
from string import ascii_uppercase, digits
from typing import Union, Optional, List, Dict
from uuid import uuid4, UUID

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError, ObjectDoesNotExist
from django.core.validators import MinValueValidator, MinLengthValidator
from django.db import models, transaction, IntegrityError
from django.db.models import Q, Sum, ExpressionWrapper, FloatField, F
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from django.urls import reverse
from django.utils.translation import gettext_lazy as _

from django_ledger.io.io_core import get_localdate
from django_ledger.models import BillModelQuerySet, InvoiceModelQuerySet, lazy_loader, deprecated_entity_slug_behavior
from django_ledger.models.customer import CustomerModel
from django_ledger.models.entity import EntityModel, EntityStateModel
from django_ledger.models.items import ItemTransactionModelQuerySet, ItemTransactionModel, ItemModelQuerySet, ItemModel
from django_ledger.models.mixins import CreateUpdateMixIn, MarkdownNotesMixIn, ItemizeMixIn
from django_ledger.models.purchase_order import PurchaseOrderModelQuerySet
from django_ledger.models.signals import (
    estimate_status_void,
    estimate_status_draft,
    estimate_status_approved,
    estimate_status_canceled,
    estimate_status_completed,
    estimate_status_in_review
)
from django_ledger.settings import DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING, DJANGO_LEDGER_ESTIMATE_NUMBER_PREFIX, \
    DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR

ESTIMATE_NUMBER_CHARS = ascii_uppercase + digits

UserModel = get_user_model()


class EstimateModelValidationError(ValidationError):
    pass


class EstimateModelQuerySet(models.QuerySet):
    """
    A custom-defined LedgerModelManager that implements custom QuerySet methods related to the EstimateModel.
    """

    def for_user(self, user_model):
        if user_model.is_superuser:
            return self
        return self.filter(
            Q(entity__admin=user_model) |
            Q(entity__managers__in=[user_model])
        )

    def approved(self):
        """
        Approved Estimates or Sales Orders are those that have been approved or completed.

        Returns
        -------
        EstimateModelQuerySet
            A EstimateModelQuerySet with applied filters.
        """
        return self.filter(
            Q(status__exact=EstimateModelAbstract.CONTRACT_STATUS_APPROVED) |
            Q(status__exact=EstimateModelAbstract.CONTRACT_STATUS_COMPLETED)
        )

    def not_approved(self):
        """
        Not approved Estimates or Sales Orders are those that have not been approved or completed.

        Returns
        -------
        EstimateModelQuerySet
            A EstimateModelQuerySet with applied filters.
        """
        return self.exclude(
            Q(status__exact=EstimateModelAbstract.CONTRACT_STATUS_APPROVED) |
            Q(status__exact=EstimateModelAbstract.CONTRACT_STATUS_COMPLETED)
        )

    def contracts(self):
        """
        A contract are Estimates or Sales Orders are those that have been approved or completed.

        Returns
        -------
        EstimateModelQuerySet
            A EstimateModelQuerySet with applied filters. Equivalent to approve.
        """
        return self.approved()

    def estimates(self):
        """
         Estimates or Sales Orders are those that have not been approved or completed.

        Returns
        -------
        EstimateModelQuerySet
            A EstimateModelQuerySet with applied filters. Equivalent to not approved.
        """
        return self.not_approved()

    def draft(self):
        return self.filter(status__exact=EstimateModelAbstract.CONTRACT_STATUS_DRAFT)


class EstimateModelManager(models.Manager):
    """
    A custom-defined EstimateModelManager that that implements custom QuerySet methods related to the EstimateModel.
    """

    @deprecated_entity_slug_behavior
    def for_entity(self, entity_model: Union[EntityModel, str, UUID] = None, **kwargs) -> EstimateModelQuerySet:
        """
        Filters the queryset based on the given entity model.

        This method allows filtering the queryset by an `entity_model` parameter, which can
        be of type `EntityModel`, `str`, or `UUID`. It also handles deprecated behavior if
        `user_model` is passed in the `kwargs`.

        Parameters
        ----------
        entity_model : EntityModel, str, UUID
            Specifies the entity against which the queryset should be filtered. Accepts an
            `EntityModel` instance, a string (representing the entity slug), or a UUID
            (representing the entity ID).
        **kwargs
            Additional keyword arguments. Supports a deprecated `user_model` argument for
            backward compatibility.

        Returns
        -------
        QuerySet
            A Django QuerySet filtered according to the given entity model.

        Raises
        ------
        EstimateModelValidationError
            If `entity_model` is of an unsupported type.
        """
        EntityModel = lazy_loader.get_entity_model()

        qs = self.get_queryset()
        if 'user_model' in kwargs:
            warnings.warn(
                'user_model parameter is deprecated and will be removed in a future release. '
                'Use for_user(user_model).for_entity(entity_model) instead to keep current behavior.',
                DeprecationWarning,
                stacklevel=2
            )
            if DJANGO_LEDGER_USE_DEPRECATED_BEHAVIOR:
                qs = qs.for_user(kwargs['user_model'])

        if isinstance(entity_model, EntityModel):
            qs = qs.filter(entity=entity_model)
        elif isinstance(entity_model, str):
            qs = qs.filter(entity__slug__exact=entity_model)
        elif isinstance(entity_model, UUID):
            qs = qs.filter(entity_id=entity_model)
        else:
            raise EstimateModelValidationError(
                message='entity_model must be either a string, UUID or an EntityModel'
            )
        return qs


class EstimateModelAbstract(CreateUpdateMixIn,
                            ItemizeMixIn,
                            MarkdownNotesMixIn):
    """
    This is the main abstract class which the EstimateModel database will inherit from.
    The EstimateModel inherits functionality from the following MixIns:

        1. :func:`MarkdownNotesMixIn <django_ledger.models.mixins.MarkdownNotesMixIn>`
        2. :func:`CreateUpdateMixIn <django_ledger.models.mixins.CreateUpdateMixIn>`

    Attributes
    ----------
    uuid : UUID
        This is a unique primary key generated for the table. The default value of this field is uuid4().
    estimate_number: str
        Auto assigned number at creation by generate_estimate_number() function.
        Prefix be customized with DJANGO_LEDGER_ESTIMATE_NUMBER_PREFIX setting.
        Includes a reference to the Fiscal Year and a sequence number. Max Length is 20.
    entity: EntityModel
        The EntityModel associated with te EntityModel instance.
    customer: CustomerModel
        The CustomerModel associated with the EstimateModel instance.
    title: str
        A string representing the name or title of the EstimateModel instance.
    status: str
        The status of the EstimateModel instance. Must be one of Draft, In Review, Approved, Completed Void or Canceled.
    terms: str
        The contract terms that will be associated with this EstimateModel instance.
        Choices are Fixed Price, Target Price, Time & Materials and Other.
    date_draft: date
        The draft date represents the date when the EstimateModel was first created. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_in_review: date
        The in review date represents the date when the EstimateModel was marked as In Review status.
        Will be null if EstimateModel is canceled during draft status. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_approved: date
        The approved date represents the date when the EstimateModel was approved.
        Will be null if EstimateModel is canceled.
        Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_completed: date
        The paid date represents the date when the EstimateModel was completed and fulfilled.
        Will be null if EstimateModel is canceled. Defaults to
        :func:`localdate <django.utils.timezone.localdate>`.
    date_void: date
        The void date represents the date when the EstimateModel was void, if applicable.
        Will be null unless EstimateModel is void. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    date_canceled: date
        The canceled date represents the date when the EstimateModel was canceled, if applicable.
        Will be null unless EstimateModel is canceled. Defaults to :func:`localdate <django.utils.timezone.localdate>`.
    revenue_estimate: Decimal
        The total estimated revenue of the EstimateModel instance.
    labor_estimate: Decimal
        The total labor costs estimate of the EstimateModel instance.
    material_estimate: Decimal
        The total material costs estimate of the EstimateModel instance.
    equipment_estimate: Decimal
        The total equipment costs estimate of the EstimateModel instance.
    other_estimate: Decimal
        the total miscellaneous costs estimate of the EstimateModel instance.
    """
    CONTRACT_STATUS_DRAFT = 'draft'
    CONTRACT_STATUS_REVIEW = 'in_review'
    CONTRACT_STATUS_APPROVED = 'approved'
    CONTRACT_STATUS_COMPLETED = 'completed'
    CONTRACT_STATUS_VOID = 'void'
    CONTRACT_STATUS_CANCELED = 'canceled'
    CONTRACT_STATUS_CHOICES = [
        (CONTRACT_STATUS_DRAFT, _('Draft')),
        (CONTRACT_STATUS_REVIEW, _('In Review')),
        (CONTRACT_STATUS_APPROVED, _('Approved')),
        (CONTRACT_STATUS_COMPLETED, _('Completed')),
        (CONTRACT_STATUS_VOID, _('Void')),
        (CONTRACT_STATUS_CANCELED, _('Canceled')),
    ]
    CONTRACT_STATUS_CHOICES_VALID = tuple(i[0] for i in CONTRACT_STATUS_CHOICES)

    CONTRACT_TERMS_FIXED = 'fixed'
    CONTRACT_TERMS_TARGET_PRICE = 'target'
    CONTRACT_TERMS_TM = 't&m'
    CONTRACT_TERMS_OTHER = 'other'
    CONTRACT_TERMS_CHOICES = [
        (CONTRACT_TERMS_FIXED, _('Fixed Price')),
        (CONTRACT_TERMS_TARGET_PRICE, _('Target Price')),
        (CONTRACT_TERMS_TM, _('Time & Materials')),
        (CONTRACT_TERMS_OTHER, _('Other'))
    ]
    CONTRACT_TERMS_CHOICES_VALID = tuple(i[0] for i in CONTRACT_TERMS_CHOICES)

    uuid = models.UUIDField(default=uuid4, editable=False, primary_key=True)
    estimate_number = models.SlugField(max_length=20,
                                       editable=False,
                                       verbose_name=_('Estimate Number'))
    entity = models.ForeignKey('django_ledger.EntityModel',
                               editable=False,
                               on_delete=models.CASCADE,
                               verbose_name=_('Entity Model'))
    customer = models.ForeignKey('django_ledger.CustomerModel', on_delete=models.RESTRICT, verbose_name=_('Customer'))
    terms = models.CharField(max_length=10, choices=CONTRACT_TERMS_CHOICES, verbose_name=_('Contract Terms'))
    title = models.CharField(max_length=250,
                             verbose_name=_('Customer Estimate Title'),
                             validators=[
                                 MinLengthValidator(limit_value=5,
                                                    message=_(f'EstimateModel Title length must be greater than 5'))
                             ])
    status = models.CharField(max_length=10,
                              choices=CONTRACT_STATUS_CHOICES,
                              verbose_name=_('Estimate Model Status'),
                              default=CONTRACT_STATUS_DRAFT)

    date_draft = models.DateField(null=True, blank=True, verbose_name=_('Date Draft'))
    date_in_review = models.DateField(null=True, blank=True, verbose_name=_('Date In Review'))
    date_approved = models.DateField(null=True, blank=True, verbose_name=_('Date Approved'))
    date_completed = models.DateField(null=True, blank=True, verbose_name=_('Date Completed'))
    date_canceled = models.DateField(null=True, blank=True, verbose_name=_('Date Canceled'))
    date_void = models.DateField(null=True, blank=True, verbose_name=_('Date Void'))

    revenue_estimate = models.DecimalField(decimal_places=2,
                                           max_digits=20,
                                           default=Decimal('0.00'),
                                           verbose_name=_('Total revenue estimate'),
                                           help_text=_('Estimated cost to complete the quoted work.'),
                                           validators=[MinValueValidator(limit_value=0)])

    labor_estimate = models.DecimalField(decimal_places=2,
                                         max_digits=20,
                                         default=Decimal('0.00'),
                                         verbose_name=_('Labor Cost of labor estimate'),
                                         help_text=_('Estimated labor cost to complete the quoted work.'),
                                         validators=[MinValueValidator(limit_value=0)])

    material_estimate = models.DecimalField(decimal_places=2,
                                            max_digits=20,
                                            default=0.0,
                                            verbose_name=_('Material Cost Estimate'),
                                            help_text=_('Estimated material cost to complete the quoted work.'),
                                            validators=[MinValueValidator(limit_value=0)])

    equipment_estimate = models.DecimalField(decimal_places=2,
                                             max_digits=20,
                                             default=Decimal('0.00'),
                                             verbose_name=_('Equipment Cost Estimate'),
                                             help_text=_('Estimated equipment cost to complete the quoted work.'),
                                             validators=[MinValueValidator(limit_value=0)])

    other_estimate = models.DecimalField(decimal_places=2,
                                         max_digits=20,
                                         default=Decimal('0.00'),
                                         verbose_name=_('Other Cost Estimate'),
                                         help_text=_('Estimated equipment cost to complete the quoted work.'),
                                         validators=[MinValueValidator(limit_value=0)])

    objects = EstimateModelManager.from_queryset(queryset_class=EstimateModelQuerySet)()

    class Meta:
        abstract = True
        ordering = ['-updated']
        verbose_name = _('Customer Job')
        verbose_name_plural = _('Customer Jobs')
        indexes = [
            models.Index(fields=['status']),
            models.Index(fields=['customer']),
            models.Index(fields=['terms']),
            models.Index(fields=['entity']),

            models.Index(fields=['date_draft']),
            models.Index(fields=['date_in_review']),
            models.Index(fields=['date_approved']),
            models.Index(fields=['date_canceled']),
            models.Index(fields=['date_void']),
            models.Index(fields=['estimate_number']),
        ]
        unique_together = [
            ('entity', 'estimate_number')
        ]

    def __str__(self):
        if self.is_contract():
            return f'Contract {self.estimate_number} | {self.title}'
        return f'Estimate {self.estimate_number} | {self.title}'

    # Configuration...

    def configure(self,
                  entity_slug: Union[EntityModel, UUID, str],
                  customer_model: CustomerModel,
                  user_model: Optional[UserModel] = None,
                  date_draft: Optional[date] = None,
                  estimate_title: Optional[str] = None,
                  commit: bool = False,
                  raise_exception: bool = True):
        """
        A configuration hook which executes all initial EstimateModel setup.
        Can only call this method once in the lifetime of a EstimateModel.

        Parameters
        ----------
        entity_slug: str or EntityModel
            The entity slug or EntityModel to associate the Bill with.
        user_model:
            The UserModel making the request to check for QuerySet permissions.
        date_draft: date
            The draft date to use. If None defaults to localdate().
        customer_model: CustomerModel
            The CustomerModel to be associated with this EstimateModel instance.
        commit: bool
            Saves the current EstimateModel after being configured.
        estimate_title: str
            Optional EstimateModel title.
        raise_exception: bool
            If True, raises EstimateModelValidationError when model is already configured.

        Returns
        -------
        EstimateModel
            The configured EstimateModel instance.
        """
        if not self.is_configured():
            if isinstance(entity_slug, (str, UUID)):
                if not user_model:
                    if raise_exception:
                        raise EstimateModelValidationError(_('Must pass user_model when using entity_slug.'))
                    return
                entity_qs = EntityModel.objects.for_user(user_model=user_model)
                if isinstance(entity_slug, str):
                    entity_model: EntityModel = get_object_or_404(entity_qs, slug__exact=entity_slug)
                elif isinstance(entity_slug, UUID):
                    entity_model: EntityModel = get_object_or_404(entity_qs, uuid__exact=entity_slug)
            elif isinstance(entity_slug, EntityModel):
                entity_model = entity_slug
            else:
                if raise_exception:
                    raise EstimateModelValidationError('entity_slug must be an instance of str or EntityModel')
                return

            if estimate_title:
                self.title = estimate_title

            self.entity = entity_model

            self.customer = customer_model
            if not date_draft:
                self.date_draft = get_localdate()

            self.clean()
            self.clean_fields()

            if commit:
                self.save()
        return self

    # State....

    def is_draft(self) -> bool:
        """
        Determines if the EstimateModel is in Draft status.

        Returns
        -------
        bool
            True if EstimateModel is in Draft status, else False.
        """
        return self.status == self.CONTRACT_STATUS_DRAFT

    def is_review(self) -> bool:
        """
        Determines if the EstimateModel is In Review status.

        Returns
        -------
        bool
            True if EstimateModel is In Review status, else False.
        """
        return self.status == self.CONTRACT_STATUS_REVIEW

    def is_approved(self) -> bool:
        """
        Determines if the EstimateModel is in Approved status.

        Returns
        -------
        bool
            True if EstimateModel is in Approved status, else False.
        """
        return self.status == self.CONTRACT_STATUS_APPROVED

    def is_completed(self) -> bool:
        """
        Determines if the EstimateModel is in Completed status.

        Returns
        -------
        bool
            True if EstimateModel is in Completed status, else False.
        """
        return self.status == self.CONTRACT_STATUS_COMPLETED

    def is_canceled(self) -> bool:
        """
        Determines if the EstimateModel is in Canceled status.

        Returns
        -------
        bool
            True if EstimateModel is in Canceled status, else False.
        """
        return self.status == self.CONTRACT_STATUS_CANCELED

    def is_void(self):
        """
        Determines if the EstimateModel is in Void status.

        Returns
        -------
        bool
            True if EstimateModel is in Void status, else False.
        """
        return self.status == self.CONTRACT_STATUS_VOID

    def is_contract(self):
        """
        Determines if the EstimateModel is considered a Contract.

        Returns
        -------
        bool
            True if EstimateModel is a Contract, else False.
        """
        return any([
            self.is_approved(),
            self.is_completed()
        ])

    def is_configured(self) -> bool:
        """
        Determines if the EstimateModel is configured.

        Returns
        -------
        bool
            True if EstimateModel is configured, else False.
        """
        return all([
            self.customer_id,
            self.entity_id,
            self.date_draft
        ])

    # Permissions...
    def can_draft(self) -> bool:
        """
        Determines if the EstimateModel can be marked as Draft.

        Returns
        -------
        bool
            True if EstimateModel can be marked as draft, else False.
        """
        return self.is_review()

    def can_review(self):
        """
        Determines if the EstimateModel can be marked as In Review.

        Returns
        -------
        bool
            True if EstimateModel can be marked as In Review, else False.
        """
        return self.is_draft()

    def can_approve(self):
        """
        Determines if the EstimateModel can be marked as approved.

        Returns
        -------
        bool
            True if EstimateModel can be marked as approved, else False.
        """
        return self.is_review()

    def can_complete(self):
        """
        Determines if the EstimateModel can be marked as completed.

        Returns
        -------
        bool
            True if EstimateModel can be marked as completed, else False.
        """
        return self.is_approved()

    def can_cancel(self):
        """
        Determines if the EstimateModel can be marked as canceled.

        Returns
        -------
        bool
            True if EstimateModel can be marked as canceled, else False.
        """
        return any([
            self.is_draft(),
            self.is_review()
        ])

    def can_void(self):
        """
        Determines if the EstimateModel can be marked as void.

        Returns
        -------
        bool
            True if EstimateModel can be marked as void, else False.
        """
        return self.is_approved()

    def can_update_items(self):
        """
        Determines if the EstimateModel item transactions can be edited or changed.

        Returns
        -------
        bool
            True if EstimateModel item transactions can be edited or changed, else False.
        """
        return self.is_draft()

    def can_bind(self):
        """
        Determines if the EstimateModel can be bound to a set of POs, Bills or Invoices.

        Returns
        -------
        bool
            True if EstimateModel can be bound, else False.
        """
        return self.is_approved()

    def can_generate_estimate_number(self):
        """
        Determines if the EstimateModel can generate its own estimate number..

        Returns
        -------
        bool
            True if EstimateModel can generate the estimate number, else False.
        """
        return all([
            self.date_draft,
            self.is_configured(),
            not self.estimate_number
        ])

    # Actions...
    # DRAFT...
    def mark_as_draft(self, commit: bool = False, raise_exception: bool = True, **kwargs):
        """
        Marks the current EstimateModel instance as Draft.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        """
        if not self.can_draft():
            if raise_exception:
                raise EstimateModelValidationError(f'Estimate {self.estimate_number} cannot be marked as draft...')
            return
        self.status = self.CONTRACT_STATUS_DRAFT
        self.clean()
        if commit:
            self.save(update_fields=[
                'status',
                'updated'
            ])
        estimate_status_draft.send_robust(sender=self.__class__,
                                          instance=self,
                                          commited=commit,
                                          **kwargs)

    def get_mark_as_draft_html_id(self):
        """
        EstimateModel Mark as Draft HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-draft'

    def get_mark_as_draft_url(self):
        """
        EstimateModel Mark as Draft URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-draft',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_draft_message(self):
        """
        EstimateModel Mark as Draft Message.

        Returns
        _______
        str
            Message as a String
        """
        return _('Do you want to mark Estimate %s as Draft?') % self.estimate_number

    # REVIEW...
    def mark_as_review(self,
                       itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None,
                       date_in_review: Optional[date] = None,
                       raise_exception: bool = True,
                       commit: bool = True,
                       **kwargs):

        """
        Marks the current EstimateModel instance as In Review.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        itemtxs_qs: ItemTransactionModelQuerySet
            Optional, pre-fetched ItemTransactionModelQuerySet. Avoids additional DB query.
            Validated if provided.
        date_in_review: date
            Optional date when EstimateModel instance is In Review. Defaults to localdate().
        """
        if not self.can_review():
            if raise_exception:
                raise ValidationError(f'Estimate {self.estimate_number} cannot be marked as In Review...')
            return

        if not itemtxs_qs:
            itemtxs_qs = self.itemtransactionmodel_set.all()
        else:
            self.validate_item_transaction_qs(itemtxs_qs=itemtxs_qs)

        if not itemtxs_qs.count():
            raise EstimateModelValidationError(message='Cannot review an Estimate without items...')
        if not self.get_cost_estimate():
            raise EstimateModelValidationError(message='Cost amount is zero!.')
        if not self.revenue_estimate:
            raise EstimateModelValidationError(message='Revenue amount is zero!.')

        if not date_in_review:
            date_in_review = get_localdate()
        self.date_in_review = date_in_review
        self.status = self.CONTRACT_STATUS_REVIEW
        self.clean()
        if commit:
            self.save(update_fields=[
                'date_in_review',
                'status',
                'updated'
            ])
        estimate_status_in_review.send_robust(sender=self.__class__,
                                              instance=self,
                                              commited=commit,
                                              **kwargs)

    def get_mark_as_review_html_id(self):
        """
        EstimateModel Mark as Review HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-review'

    def get_mark_as_review_url(self):
        """
        EstimateModel Mark as In Review URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-review',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_review_message(self):
        """
        EstimateModel Mark as In Review Message.

        Returns
        -------
        str
            Message as a String
        """
        return _('Do you want to mark Estimate %s as In Review?') % self.estimate_number

    # APPROVED
    def mark_as_approved(self,
                         commit=False,
                         date_approved: Optional[date] = None,
                         raise_exception: bool = True,
                         **kwargs):
        """
        Marks the current EstimateModel instance as Approved.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        date_approved: date
            Optional date when EstimateModel instance is Approved. Defaults to localdate().
        """
        if not self.can_approve():
            if raise_exception:
                raise EstimateModelValidationError(
                    f'Estimate {self.estimate_number} cannot be marked as approved.'
                )
            return

        if not date_approved:
            date_approved = get_localdate()
        self.date_approved = date_approved
        self.status = self.CONTRACT_STATUS_APPROVED
        self.clean()
        if commit:
            self.save(update_fields=[
                'status',
                'date_approved',
                'updated'
            ])
        estimate_status_approved.send_robust(sender=self.__class__,
                                             instance=self,
                                             commited=commit,
                                             **kwargs)

    def get_mark_as_approved_html_id(self):
        """
        EstimateModel Mark as Approved HTML ID Tag.

        Returns
        -------
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-approved'

    def get_mark_as_approved_url(self):
        """
        EstimateModel Mark as Approved URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-approved',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_approved_message(self):
        """
        EstimateModel Mark as Approved Message.

        Returns
        _______
        str
            Message as a String
        """
        return _('Do you want to mark Estimate %s as Approved?') % self.estimate_number

    # COMPLETED
    def mark_as_completed(self,
                          commit=False,
                          date_completed: Optional[date] = None,
                          raise_exception: bool = True,
                          **kwargs):
        """
        Marks the current EstimateModel instance as Completed.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        date_completed: date
            Optional date when EstimateModel instance is completed. Defaults to localdate().
        """
        if not self.can_complete():
            if raise_exception:
                raise EstimateModelValidationError(f'Estimate {self.estimate_number} cannot be marked as completed.')
            return
        if not date_completed:
            date_completed = get_localdate()
        self.date_completed = date_completed
        self.status = self.CONTRACT_STATUS_COMPLETED
        self.clean()
        if commit:
            self.clean()
            self.save(
                update_fields=[
                    'status',
                    'date_completed',
                    'updated'
                ])
        estimate_status_completed.send_robust(
            sender=self.__class__,
            instance=self,
            commited=commit,
            **kwargs
        )

    def get_mark_as_completed_html_id(self):
        """
        EstimateModel Mark as Complete HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-completed'

    def get_mark_as_completed_url(self):
        """
        EstimateModel Mark as Completed URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-completed',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_completed_message(self):
        """
        EstimateModel Mark as Completed Message.

        Returns
        _______
        str
            Message as a String
        """
        return _('Do you want to mark Estimate %s as Completed?') % self.estimate_number

    # CANCEL
    def mark_as_canceled(self,
                         commit: bool = False,
                         date_canceled: Optional[date] = None,
                         raise_exception: bool = True,
                         **kwargs):
        """
        Marks the current EstimateModel instance as Canceled.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        date_canceled: date
            Optional date when EstimateModel instance is canceled. Defaults to localdate().
        """
        if not self.can_cancel():
            if raise_exception:
                raise EstimateModelValidationError(f'Estimate {self.estimate_number} cannot be canceled...')
            return
        if not date_canceled:
            date_canceled = get_localdate()
        self.date_canceled = date_canceled
        self.status = self.CONTRACT_STATUS_CANCELED
        self.clean()
        if commit:
            self.save(update_fields=[
                'status',
                'date_canceled',
                'updated'
            ])
        estimate_status_canceled.send_robust(
            sender=self.__class__,
            instance=self,
            commited=commit,
            **kwargs
        )

    def get_mark_as_canceled_html_id(self):
        """
        EstimateModel Mark as Canceled HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-canceled'

    def get_mark_as_canceled_url(self):
        """
        EstimateModel Mark as Canceled URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-canceled',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_canceled_message(self):
        """
        EstimateModel Mark as Canceled Message.

        Returns
        _______
        str
            Message as a String
        """
        return _('Do you want to mark Estimate %s as Canceled?') % self.estimate_number

    # VOID
    def mark_as_void(self,
                     commit: bool = False,
                     date_void: Optional[date] = None,
                     raise_exception: bool = True,
                     **kwargs):

        """
        Marks the current EstimateModel instance as Void.

        Parameters
        ----------
        commit: bool
            Commits transaction into current EstimateModel instance.
        date_void: date
            Optional date when EstimateModel instance is void. Defaults to localdate().
        """
        if not self.can_void():
            if raise_exception:
                raise EstimateModelValidationError(f'Estimate {self.estimate_number} cannot be void...')
            return

        if not date_void:
            date_void = get_localdate()
        self.date_void = date_void
        self.status = self.CONTRACT_STATUS_VOID
        self.clean()
        if commit:
            self.save(update_fields=[
                'status',
                'date_void',
                'updated'
            ])
        estimate_status_void.send_robust(
            sender=self.__class__,
            instance=self,
            commited=commit,
            **kwargs
        )

    def get_mark_as_void_html_id(self):
        """
        EstimateModel Mark as Void HTML ID Tag.

        Returns
        _______
        str
            HTML ID as a String
        """
        return f'djl-{self.uuid}-estimate-mark-as-void'

    def get_mark_as_void_url(self):
        """
        EstimateModel Mark as Void URL.

        Returns
        _______
        str
            URL as a String
        """
        return reverse('django_ledger:customer-estimate-action-mark-as-void',
                       kwargs={
                           'entity_slug': self.entity.slug,
                           'ce_pk': self.uuid
                       })

    def get_mark_as_void_message(self):
        """
        EstimateModel Mark as Void message.

        Returns
        _______
        str
            Message as a String
        """

        return _('Do you want to mark Estimate %s as Void?') % self.estimate_number

    # HTML Tags...
    def get_html_id(self):
        return f'djl-customer-estimate-id-{self.uuid}'

    # ### ItemizeMixIn implementation START...
    def can_migrate_itemtxs(self) -> bool:
        return self.is_draft()

    def migrate_itemtxs(self, itemtxs: Dict, operation: str, commit: bool = False):
        itemtxs_batch = super().migrate_itemtxs(itemtxs=itemtxs, commit=commit, operation=operation)
        self.update_state(itemtxs_qs=itemtxs_batch)
        self.clean()

        if commit:
            self.save(update_fields=[
                'revenue_estimate',
                'labor_estimate',
                'equipment_estimate',
                'material_estimate',
                'other_estimate',
                'updated'
            ])

        return itemtxs_batch

    def get_item_model_qs(self) -> ItemModelQuerySet:
        return ItemModel.objects.filter(
            entity_id__exact=self.entity_id
        ).estimates()

    def validate_itemtxs_qs(self, queryset: Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]):
        """
        Validates that the entire ItemTransactionModelQuerySet is bound to the EstimateModel.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet or list of ItemTransactionModel.
            ItemTransactionModelQuerySet to validate.
        """
        valid = all([
            i.ce_model_id == self.uuid for i in queryset
        ])
        if not valid:
            raise EstimateModelValidationError(f'Invalid queryset. All items must be assigned to Bill {self.uuid}')

    def get_itemtxs_data(self,
                         queryset: Optional[Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]] = None,
                         aggregate_on_db: bool = False,
                         lazy_agg: bool = False):

        """
        Returns all ItemTransactionModels associated with the EstimateModel and a total aggregate.

        Parameters
        ----------
        queryset: ItemTransactionModelQuerySet
            ItemTransactionModelQuerySet to use. Avoids additional DB query if provided.
            Validated if provided.

        Returns
        -------
        ItemTransactionModelQuerySet
        """
        if not queryset:
            queryset = self.itemtransactionmodel_set.select_related('item_model').all()
        else:
            self.validate_item_transaction_qs(queryset)
        # todo: this needs to return an aggregate for consistency...
        return queryset, None

    # ### ItemizeMixIn implementation END...
    def get_itemtxs_annotation(self, itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None):
        """
        Gets an annotated ItemTransactionModelQuerySet with additional average unit cost & revenue.

        Parameters
        ----------

        itemtxs_qs: ItemTransactionModelQuerySet
            Prefetched ItemTransactionModelQuerySet if any. If None a new queryset will be evaluated.
            Will be validated if provided.

        Returns
        -------
        tuple
            The original ItemTransactionModelQuerySet and the annotated ItemTransactionModelQuerySet.
        """
        itemtxs_qs, _ = self.get_itemtxs_data(itemtxs_qs)
        return itemtxs_qs, itemtxs_qs.values(
            'item_model_id', 'item_model__name'
        ).annotate(
            Sum('ce_quantity'),
            Sum('ce_cost_estimate'),
            Sum('ce_revenue_estimate'),
            avg_unit_cost=ExpressionWrapper(
                expression=Sum('ce_cost_estimate') / Sum('ce_quantity'),
                output_field=FloatField()
            ),
            avg_unit_revenue=ExpressionWrapper(
                expression=Sum('ce_revenue_estimate') / Sum('ce_quantity'),
                output_field=FloatField()
            )
        )

    def update_revenue_estimate(self, itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None, commit: bool = False):
        """
        Updates the revenue estimate of the EstimateModel instance.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet
            Prefetched ItemTransactionModelQuerySet. A new ItemTransactionModelQuerySet will be fetched from DB if not
            provided. If provided will be validated.

        commit: bool
            If True, the new revenue estimate will be committed into the DB.
        """
        itemtxs_qs, _ = self.get_itemtxs_data(itemtxs_qs)
        self.revenue_estimate = sum(i.ce_revenue_estimate for i in itemtxs_qs)

        if commit:
            self.save(update_fields=[
                'revenue_estimate',
                'updated'
            ])

    def update_cost_estimate(self, itemtxs_qs: Optional[ItemTransactionModelQuerySet] = None, commit: bool = False):
        """
        Updates the cost estimate of the EstimateModel instance.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet
            Prefetched ItemTransactionModelQuerySet. A new ItemTransactionModelQuerySet will be fetched from DB if not
            provided. If provided will be validated.
        commit: bool
            If True, the new revenue estimate will be committed into the DB.
        """
        itemtxs_qs, _ = self.get_itemtxs_data(queryset=itemtxs_qs)
        estimates = {
            'labor': sum(a.ce_cost_estimate for a in itemtxs_qs if a.item_model.is_labor()),
            'material': sum(a.ce_cost_estimate for a in itemtxs_qs if a.item_model.is_material()),
            'equipment': sum(a.ce_cost_estimate for a in itemtxs_qs if a.item_model.is_equipment()),
            'other': sum(
                a.ce_cost_estimate for a in itemtxs_qs
                if
                a.item_model.is_other() or not a.item_model_id or not a.item_model.item_type or a.item_model.is_lump_sum()
            ),
        }
        self.labor_estimate = estimates['labor']
        self.material_estimate = estimates['material']
        self.equipment_estimate = estimates['equipment']
        self.other_estimate = estimates['other']

        if commit:
            self.save(update_fields=[
                'labor_estimate',
                'material_estimate',
                'equipment_estimate',
                'other_estimate',
                'updated'
            ])

    def update_state(self,
                     itemtxs_qs: Optional[Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]] = None):
        itemtxs_qs, _ = self.get_itemtxs_data(queryset=itemtxs_qs)
        self.update_cost_estimate(itemtxs_qs)
        self.update_revenue_estimate(itemtxs_qs)

    # Features...
    def get_cost_estimate(self, as_float: bool = False) -> Union[float, Decimal]:
        """
        Calculates the total EstimateModel cost by summing all individual cost components: Labor, Material, Equipment &
        Other costs.

        Parameters
        ----------
        as_float: bool
            If True, will return value as float, else as Decimal.

        Returns
        -------
        Decimal or float
            The total EstimateModel cost.
        """
        estimate = sum([
            self.labor_estimate,
            self.material_estimate,
            self.equipment_estimate,
            self.other_estimate
        ])
        if as_float:
            return float(estimate)
        return estimate

    def get_revenue_estimate(self, as_float: bool = False):
        """
        Calculates the total EstimateModel revenue.

        Parameters
        ----------
        as_float: bool
            If True, will return value as float, else as Decimal.

        Returns
        -------
        Decimal or float
            The total EstimateModel revenue.
        """
        if as_float:
            return float(self.revenue_estimate)
        return self.revenue_estimate

    def get_profit_estimate(self, as_float: bool = False):
        """
        Calculates the total EstimateModel profit estimate.

        Parameters
        ----------
        as_float: bool
            If True, will return value as float, else as Decimal.

        Returns
        -------
        Decimal or float
            The total EstimateModel profit.
        """

        return self.get_revenue_estimate(as_float=as_float) - self.get_cost_estimate(as_float=as_float)

    def get_gross_margin_estimate(self, as_percent: bool = False, raise_exception: bool = False) -> float:
        """
        Computes the EstimateModel gross margin.

        Parameters
        ----------
        as_percent: bool
            If True, computes gross margin as percent.

        raise_exception: bool

        Returns
        -------
        float
            The EstimateModel gross margin.
        """
        try:
            gm = float(self.get_revenue_estimate()) / float(self.get_cost_estimate()) - 1.00
            if as_percent:
                return gm * 100
            return gm
        except ZeroDivisionError as e:
            if raise_exception:
                raise EstimateModelValidationError(message=_('Cannot compute gross margin, total cost is zero.'))
            return 0.00

    def get_status_action_date(self) -> date:
        """
        The date when the latest status took place.

        Examples
        --------
        If EstimateModel is Draft, will return date_draft.
        If EstimateModel is In Review, will return date_in_review.
        If EstimateModel is Approved, will return date_approved.

        Returns
        -------
        date
            The latest action date.
        """
        return getattr(self, f'date_{self.status}')

    # --- CONTRACT METHODS ---

    # Queryset validation....
    def validate_item_transaction_qs(self, itemtxs_qs: Union[ItemTransactionModelQuerySet, List[ItemTransactionModel]]):
        """
        Validates that the entire ItemTransactionModelQuerySet is bound to the BillModel.

        Parameters
        ----------
        itemtxs_qs: ItemTransactionModelQuerySet
            ItemTransactionModelQuerySet to validate.
        """
        if not isinstance(itemtxs_qs, ItemTransactionModelQuerySet):
            if not all([
                isinstance(i, ItemTransactionModel) for i in itemtxs_qs
            ]):
                raise EstimateModelValidationError(
                    message='Must pass an instance of ItemTransactionModelQuerySet or a list of ItemTransactionModel'
                )
        valid = all([
            i.ce_model_id == self.uuid for i in itemtxs_qs
        ])
        if not valid:
            raise EstimateModelValidationError(f'Invalid queryset. All items must be assigned to Estimate {self.uuid}')

    def validate_po_queryset(self, po_qs: PurchaseOrderModelQuerySet) -> PurchaseOrderModelQuerySet:
        """
        Validates a prefetched PurchaseOrderModelQuerySet against the EstimateModel instance.

        Parameters
        ----------
        po_qs: PurchaseOrderModelQuerySet
            The PurchaseOrderModelQuerySet to validate

        Returns
        -------
        PurchaseOrderModelQuerySet
            The validated PurchaseOrderModelQuerySet.
        """
        valid = all([
            po_model.ce_model_id == self.uuid for po_model in po_qs
        ])
        if not valid:
            raise EstimateModelValidationError(message='Invalid PurchaseOrderModelQuerySet provided. All POs must be '
                                                       f'assigned to EstimateModel {self.uuid}')
        return po_qs

    def validate_bill_queryset(self, bill_qs: BillModelQuerySet) -> BillModelQuerySet:
        """
        Validates a prefetched BillModelQuerySet against the EstimateModel instance.

        Parameters
        ----------
        bill_qs: BillModelQuerySet
            The BillModelQuerySet to validate

        Returns
        -------
        BillModelQuerySet
            The validated BillModelQuerySet.
        """
        valid = all([
            bill_model.ce_model_id == self.uuid for bill_model in bill_qs
        ])
        if not valid:
            raise EstimateModelValidationError(message='Invalid BillModelQuerySet provided. All bills must be '
                                                       f'assigned to EstimateModel {self.uuid}')
        return bill_qs

    def validate_invoice_queryset(self, invoice_qs: InvoiceModelQuerySet) -> InvoiceModelQuerySet:
        """
        Validates a prefetched InvoiceModelQuerySet against the EstimateModel instance.

        Parameters
        ----------
        invoice_qs: InvoiceModelQuerySet
            The InvoiceModelQuerySet to validate

        Returns
        -------
        InvoiceModelQuerySet
            The validated InvoiceModelQuerySet.
        """
        valid = all([
            invoice_model.ce_model_id == self.uuid for invoice_model in invoice_qs
        ])
        if not valid:
            raise EstimateModelValidationError(message='Invalid InvoiceModelQuerySet provided. All invoices must be '
                                                       f'assigned to EstimateModel {self.uuid}')
        return invoice_qs

    def get_po_amount(self, po_qs: PurchaseOrderModelQuerySet = None) -> dict:
        if not po_qs:
            po_qs = self.purchaseordermodel_set.all().active()
        else:
            po_qs = self.validate_po_queryset(po_qs=po_qs)

        return po_qs.aggregate(po_amount__sum=Coalesce(Sum('po_amount'), 0.0, output_field=models.FloatField()))

    def get_billed_amount(self, bill_qs: Optional[BillModelQuerySet] = None) -> dict:
        if not bill_qs:
            bill_qs = self.billmodel_set.all().active()
        else:
            bill_qs = self.validate_bill_queryset(bill_qs=bill_qs)

        return bill_qs.aggregate(
            bill_amount_due__sum=Coalesce(Sum('amount_due'), 0.0, output_field=models.FloatField()),
            bill_amount_paid__sum=Coalesce(Sum('amount_paid'), 0.0, output_field=models.FloatField()),
            bill_amount_receivable__sum=Coalesce(Sum('amount_receivable'), 0.0, output_field=models.FloatField()),
            bill_amount_earned__sum=Coalesce(Sum('amount_earned'), 0.0, output_field=models.FloatField()),
            bill_amount_unearned__sum=Coalesce(Sum('amount_unearned'), 0.0, output_field=models.FloatField()),
        )

    def get_invoiced_amount(self, invoice_qs: Optional[InvoiceModelQuerySet] = None) -> dict:
        if not invoice_qs:
            invoice_qs = self.invoicemodel_set.all().active()
        else:
            invoice_qs = self.validate_invoice_queryset(invoice_qs=invoice_qs)

        return invoice_qs.aggregate(
            invoice_amount_due__sum=Coalesce(Sum('amount_due'), 0.0, output_field=models.FloatField()),
            invoice_amount_paid__sum=Coalesce(Sum('amount_paid'), 0.0, output_field=models.FloatField()),
            invoice_amount_receivable__sum=Coalesce(Sum('amount_receivable'), 0.0, output_field=models.FloatField()),
            invoice_amount_earned__sum=Coalesce(Sum('amount_earned'), 0.0, output_field=models.FloatField()),
            invoice_amount_unearned__sum=Coalesce(Sum('amount_unearned'), 0.0, output_field=models.FloatField()),
        )

    def get_contract_summary(self,
                             po_qs: Optional[PurchaseOrderModelQuerySet] = None,
                             bill_qs: Optional[BillModelQuerySet] = None,
                             invoice_qs: Optional[InvoiceModelQuerySet] = None) -> dict:
        """
        Computes an aggregate of all related ItemTransactionModels summarizing
        original contract amounts, amounts authorized, amounts billed and amount invoiced.

        Parameters
        ----------
        po_qs: PurchaseOrderModelQuerySet
            Prefetched PurchaseOrderModelQuerySet. If provided will be validated.
        bill_qs: BillModelQuerySet
            Prefetched BillModelQuerySet. If provided will be validated.
        invoice_qs: InvoiceModelQuerySet
            Prefetched InvoiceModelQuerySet. If provided will be validated.

        Returns
        -------
        dict
            A dictionary of aggregated values.
        """
        if not self.is_contract():
            raise EstimateModelValidationError(message=f'EstimateModel {self.uuid} is not a Contract.')

        stats = {
            'cost_estimate': self.get_cost_estimate(as_float=True),
            'revenue_estimate': self.get_revenue_estimate(as_float=True)
        }

        po_status = self.get_po_amount(po_qs)
        billing_status = self.get_billed_amount(bill_qs)
        invoice_status = self.get_invoiced_amount(invoice_qs)
        return stats | po_status | billing_status | invoice_status

    def _get_next_state_model(self, raise_exception: bool = True):
        """
        Fetches the next sequenced state model associated with the EstimateModel number.

        Parameters
        ----------
        raise_exception: bool
            Raises IntegrityError if unable to secure transaction from DB.

        Returns
        -------
        EntityStateModel
            An instance of EntityStateModel
        """
        # entity_model = EntityModel.objects.get(uuid__exact=self.entity_id)
        entity_model: EntityModel = self.entity
        fy_key = entity_model.get_fy_for_date(dt=self.date_draft)
        try:
            LOOKUP = {
                'entity_model_id__exact': self.entity_id,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key__exact': EntityStateModel.KEY_ESTIMATE
            }

            state_model_qs = EntityStateModel.objects.filter(**LOOKUP).select_related(
                'entity_model').select_for_update()
            state_model = state_model_qs.get()
            state_model.sequence = F('sequence') + 1
            state_model.save()
            state_model.refresh_from_db()
            return state_model
        except ObjectDoesNotExist:
            entity_model = EntityModel.objects.get(uuid__exact=self.entity_id)
            fy_key = entity_model.get_fy_for_date(dt=self.date_draft)

            LOOKUP = {
                'entity_model_id': entity_model.uuid,
                'entity_unit_id': None,
                'fiscal_year': fy_key,
                'key': EntityStateModel.KEY_ESTIMATE,
                'sequence': 1
            }

            state_model = EntityStateModel.objects.create(**LOOKUP)
            return state_model
        except IntegrityError as e:
            if raise_exception:
                raise e

    def generate_estimate_number(self, commit: bool = False) -> str:
        """
        Atomic Transaction. Generates the next EstimateModel document number available. The operation
        will result in two additional queries if the EntityModel is not cached in
        QuerySet via select_related('entity').

        Parameters
        __________
        commit: bool
            Commits transaction into BillModel.

        Returns
        _______
        str
            A String, representing the generated BillModel instance Document Number.
        """
        if self.can_generate_estimate_number():
            with transaction.atomic(durable=True):

                state_model = None
                while not state_model:
                    state_model = self._get_next_state_model(raise_exception=False)

                seq = str(state_model.sequence).zfill(DJANGO_LEDGER_DOCUMENT_NUMBER_PADDING)
                self.estimate_number = f'{DJANGO_LEDGER_ESTIMATE_NUMBER_PREFIX}-{state_model.fiscal_year}-{seq}'

                if commit:
                    self.save(update_fields=['estimate_number'])

        return self.estimate_number

    def clean(self):

        if not self.date_draft:
            self.date_draft = get_localdate()

        if self.can_generate_estimate_number():
            self.generate_estimate_number(commit=False)

        if self.is_approved() and not self.date_approved:
            self.date_approved = get_localdate()

        if self.is_completed() and not self.date_completed:
            self.date_completed = get_localdate()

        if self.is_canceled():
            self.date_approved = None
            self.date_completed = None

    def save(self, **kwargs):
        if self.can_generate_estimate_number():
            self.generate_estimate_number(commit=False)
        super(EstimateModelAbstract, self).save(**kwargs)


class EstimateModel(EstimateModelAbstract):
    """
    Base EstimateModel Class.
    """

    class Meta(EstimateModelAbstract.Meta):
        abstract = False



"""
Django Ledger created by Miguel Sanda <msanda@arrobalytics.com>.
Copyright© EDMA Group Inc licensed under the GPLv3 Agreement.

Contributions to this module:
Miguel Sanda <msanda@arrobalytics.com>
"""

from django import forms
from django.forms import ModelForm, Select, TextInput, BaseModelFormSet, modelformset_factory, Textarea
from django.utils.translation import gettext_lazy as _

from django_ledger.models import CustomerModel, ItemTransactionModel, ItemModel, EntityUnitModel
from django_ledger.models.estimate import EstimateModel
from django_ledger.settings import DJANGO_LEDGER_FORM_INPUT_CLASSES


class EstimateModelCreateForm(forms.ModelForm):

    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.ENTITY_SLUG = entity_slug
        self.USER_MODEL = user_model
        self.fields['customer'].queryset = self.get_customer_queryset()

    def get_customer_queryset(self):
        return CustomerModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG,
        ).active()

    class Meta:
        model = EstimateModel
        fields = ['title', 'customer', 'terms']
        widgets = {
            'customer': forms.Select(attrs={
                'id': 'djl-customer-estimate-customer-input',
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'terms': forms.Select(attrs={
                'id': 'djl-customer-estimate-terms-input',
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES
            }),
            'title': forms.TextInput(attrs={
                'id': 'djl-customer-job-title-input',
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-large',
                'placeholder': _('Estimate title...')
            })
        }


class BaseEstimateModelUpdateForm(forms.ModelForm):

    def __init__(self, *args, entity_slug, user_model, **kwargs):
        super().__init__(*args, **kwargs)
        self.USER_MODEL = user_model
        self.ENTITY_SLUG = entity_slug
        self.CUSTOMER_ESTIMATE_MODEL: EstimateModel = self.instance

    class Meta:
        model = EstimateModel
        fields = [
            'markdown_notes'
        ]
        widgets = {
            'terms': Select(attrs={'class': DJANGO_LEDGER_FORM_INPUT_CLASSES}),
            'markdown_notes': Textarea(attrs={
                'class': 'textarea'
            })
        }


class DraftEstimateModelUpdateForm(BaseEstimateModelUpdateForm):
    class Meta(BaseEstimateModelUpdateForm.Meta):
        fields = [
            'terms',
            'markdown_notes'
        ]


class EstimateItemModelForm(ModelForm):
    class Meta:
        model = ItemTransactionModel
        fields = [
            'item_model',
            'entity_unit',
            'ce_quantity',
            'ce_unit_cost_estimate',
            'ce_unit_revenue_estimate',
        ]
        widgets = {
            'item_model': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'entity_unit': Select(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'ce_unit_cost_estimate': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'ce_unit_revenue_estimate': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            }),
            'ce_quantity': TextInput(attrs={
                'class': DJANGO_LEDGER_FORM_INPUT_CLASSES + ' is-small',
            })
        }


class BaseEstimateItemModelFormset(BaseModelFormSet):

    def __init__(self, *args, entity_slug, user_model, customer_job_model: EstimateModel, **kwargs):
        super().__init__(*args, **kwargs)
        self.USER_MODEL = user_model
        self.ESTIMATE_MODEL = customer_job_model
        self.ENTITY_SLUG = entity_slug

        items_qs = ItemModel.objects.for_estimate(
            entity_model=self.ENTITY_SLUG
        )

        unit_qs = EntityUnitModel.objects.for_entity(
            entity_model=self.ENTITY_SLUG
        )

        for form in self.forms:
            form.fields['item_model'].queryset = items_qs
            form.fields['entity_unit'].queryset = unit_qs

            if not self.ESTIMATE_MODEL.can_update_items():
                form.fields['item_model'].disabled = True
                form.fields['ce_unit_cost_estimate'].disabled = True
                form.fields['ce_quantity'].disabled = True
                form.fields['ce_unit_revenue_estimate'].disabled = True
                form.fields['entity_unit'].disabled = True


CanEditEstimateItemModelFormset = modelformset_factory(
    model=ItemTransactionModel,
    form=EstimateItemModelForm,
    formset=BaseEstimateItemModelFormset,
    can_delete=True,
    extra=5
)

ReadOnlyEstimateItemModelFormset = modelformset_factory(
    model=ItemTransactionModel,
    form=EstimateItemModelForm,
    formset=BaseEstimateItemModelFormset,
    can_delete=False,
    extra=0
)
